#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import time
import os
import traceback

from .functions import (setLocalPath, showCommitDialog, logInPavlovia,
                        noGitWarning)
from psychopy.localization import _translate
from psychopy.projects import pavlovia
from psychopy import logging

from psychopy.app.pavlovia_ui import sync

import wx
from wx.lib import scrolledpanel as scrlpanel

try:
    import wx.lib.agw.hyperlink as wxhl  # 4.0+
except ImportError:
    import wx.lib.hyperlink as wxhl  # <3.0.2


class ProjectEditor(wx.Dialog):
    def __init__(self, parent=None, id=wx.ID_ANY, project=None, localRoot="",
                 *args, **kwargs):

        wx.Dialog.__init__(self, parent, id,
                           *args, **kwargs)
        panel = wx.Panel(self, wx.ID_ANY, style=wx.TAB_TRAVERSAL)
        # when a project is successfully created these will be populated
        if hasattr(parent, 'filename'):
            self.filename = parent.filename
        else:
            self.filename = None
        self.project = project  # type: PavloviaProject
        self.projInfo = None
        self.parent = parent

        if project:
            # edit existing project
            self.isNew = False
            if project.localRoot and not localRoot:
                localRoot = project.localRoot
        else:
            self.isNew = True

        # create the controls
        nameLabel = wx.StaticText(panel, -1, _translate("Name:"))
        self.nameBox = wx.TextCtrl(panel, -1, size=(400, -1))
        # Path can contain only letters, digits, '_', '-' and '.'.
        # Cannot start with '-', end in '.git' or end in '.atom']
        pavSession = pavlovia.getCurrentSession()

        try:
            username = pavSession.user.username
        except AttributeError as e:
            raise pavlovia.NoUserError("{}: Tried to create project with no user logged in.".format(e))

        gpChoices = [username]
        gpChoices.extend(pavSession.listUserGroups())
        groupLabel = wx.StaticText(panel, -1, _translate("Group/owner:"))
        self.groupBox = wx.Choice(panel, -1, size=(400, -1),
                                  choices=gpChoices)

        descrLabel = wx.StaticText(panel, -1, _translate("Description:"))
        self.descrBox = wx.TextCtrl(panel, -1, size=(400, 200),
                                    style=wx.TE_MULTILINE | wx.SUNKEN_BORDER)

        localLabel = wx.StaticText(panel, -1, _translate("Local folder:"))
        self.localBox = wx.TextCtrl(panel, -1, size=(400, -1),
                                    value=localRoot)
        self.btnLocalBrowse = wx.Button(panel, wx.ID_ANY, _translate("Browse..."))
        self.btnLocalBrowse.Bind(wx.EVT_BUTTON, self.onBrowseLocal)
        localPathSizer = wx.BoxSizer(wx.HORIZONTAL)
        localPathSizer.Add(self.localBox)
        localPathSizer.Add(self.btnLocalBrowse)

        tagsLabel = wx.StaticText(panel, -1,
                                  _translate("Tags (comma separated):"))
        self.tagsBox = wx.TextCtrl(panel, -1, size=(400, 100),
                                   value="PsychoPy, Builder, Coder",
                                   style=wx.TE_MULTILINE | wx.SUNKEN_BORDER)
        publicLabel = wx.StaticText(panel, -1, _translate("Public:"))
        self.publicBox = wx.CheckBox(panel, -1)

        # buttons
        if self.isNew:
            buttonMsg = _translate("Create project on Pavlovia")
        else:
            buttonMsg = _translate("Submit changes to Pavlovia")
        updateBtn = wx.Button(panel, -1, buttonMsg)
        updateBtn.Bind(wx.EVT_BUTTON, self.submitChanges)
        cancelBtn = wx.Button(panel, -1, _translate("Cancel"))
        cancelBtn.Bind(wx.EVT_BUTTON, self.onCancel)
        btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        btnSizer.AddMany([updateBtn, cancelBtn])

        # do layout
        fieldsSizer = wx.FlexGridSizer(cols=2, rows=6, vgap=5, hgap=5)
        fieldsSizer.AddMany([(nameLabel, 0, wx.ALIGN_RIGHT), self.nameBox,
                             (groupLabel, 0, wx.ALIGN_RIGHT), self.groupBox,
                             (localLabel, 0, wx.ALIGN_RIGHT), localPathSizer,
                             (descrLabel, 0, wx.ALIGN_RIGHT), self.descrBox,
                             (tagsLabel, 0, wx.ALIGN_RIGHT), self.tagsBox,
                             (publicLabel, 0, wx.ALIGN_RIGHT), self.publicBox])

        border = wx.BoxSizer(wx.VERTICAL)
        border.Add(fieldsSizer, 0, wx.ALL, 5)
        border.Add(btnSizer, 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        panel.SetSizerAndFit(border)
        self.Fit()

    def onCancel(self, evt=None):
        self.EndModal(wx.ID_CANCEL)

    def submitChanges(self, evt=None):
        session = pavlovia.getCurrentSession()
        if not session.user:
            user = logInPavlovia(parent=self.parent)
        if not session.user:
            return
        # get current values
        name = self.nameBox.GetValue()
        namespace = self.groupBox.GetStringSelection()
        descr = self.descrBox.GetValue()
        visibility = self.publicBox.GetValue()
        # tags need splitting and then
        tagsList = self.tagsBox.GetValue().split(',')
        tags = [thisTag.strip() for thisTag in tagsList]
        localRoot = self.localBox.GetValue()
        if not localRoot:
            localRoot = setLocalPath(self.parent, project=None, path="")

        # then create/update
        if self.isNew:
            project = session.createProject(name=name,
                                            description=descr,
                                            tags=tags,
                                            visibility=visibility,
                                            localRoot=localRoot,
                                            namespace=namespace)
            self.project = project
            self.project._newRemote = True
        else:  # we're changing metadata of an existing project. Don't sync
            self.project.pavlovia.name = name
            self.project.pavlovia.description = descr
            self.project.tags = tags
            self.project.visibility = visibility
            self.project.localRoot = localRoot
            self.project.save()  # pushes changed metadata to gitlab
            self.project._newRemote = False

        self.EndModal(wx.ID_OK)
        pavlovia.knownProjects.save()
        self.project.getRepo(forceRefresh=True)
        self.parent.project = self.project

    def onBrowseLocal(self, evt=None):
        newPath = setLocalPath(self, path=self.filename)
        if newPath:
            self.localBox.SetLabel(newPath)
            self.Layout()
            if self.project:
                self.project.localRoot = newPath
        self.Raise()


class DetailsPanel(scrlpanel.ScrolledPanel):

    def __init__(self, parent, noTitle=False,
                 style=wx.VSCROLL | wx.NO_BORDER,
                 project={}):

        scrlpanel.ScrolledPanel.__init__(self, parent, -1, style=style)
        self.parent = parent
        self.project = project  # type: PavloviaProject
        self.noTitle = noTitle
        self.localFolder = ''
        self.syncPanel = None

        if not noTitle:
            self.title = wx.StaticText(parent=self, id=-1,
                                       label="", style=wx.ALIGN_CENTER)
            font = wx.Font(18, wx.DECORATIVE, wx.NORMAL, wx.BOLD)
            self.title.SetFont(font)

        # if we've synced before we should know the local location
        self.localFolderCtrl = wx.StaticText(
            parent=self, id=wx.ID_ANY,
            label=_translate("Local root: "))
        self.browseLocalBtn = wx.Button(parent=self, id=wx.ID_ANY,
                                        label=_translate("Browse..."))
        self.browseLocalBtn.Bind(wx.EVT_BUTTON, self.onBrowseLocalFolder)

        # remote attributes
        self.url = wxhl.HyperLinkCtrl(parent=self, id=-1,
                                      label="https://pavlovia.org",
                                      URL="https://pavlovia.org",
                                      )
        self.description = wx.StaticText(parent=self, id=-1,
                                         label=_translate(
                                             "Select a project for details"))
        self.tags = wx.StaticText(parent=self, id=-1,
                                  label="")
        self.visibility = wx.StaticText(parent=self, id=-1,
                                        label="")

        self.syncButton = wx.Button(self, -1, _translate("Sync..."))
        self.syncButton.Enable(False)
        self.syncButton.Bind(wx.EVT_BUTTON, self.onSyncButton)
        self.syncPanel = sync.SyncStatusPanel(parent=self, id=wx.ID_ANY)

        # layout
        # sizers: on the right we have detail
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        # self.sizer.Add(wx.StaticText(self, -1, _translate("Project Info")),
        #                flag=wx.ALL,
        #                border=5)
        if not noTitle:
            self.sizer.Add(self.title, border=5,
                           flag=wx.ALL | wx.CENTER)
        self.sizer.Add(self.url, border=5,
                       flag=wx.ALL | wx.CENTER)
        self.sizer.Add(self.localFolderCtrl, border=5,
                             flag=wx.ALL | wx.EXPAND),
        self.sizer.Add(self.browseLocalBtn, border=5,
                             flag=wx.ALL | wx.LEFT)
        self.sizer.Add(self.tags, border=5, flag=wx.ALL | wx.EXPAND)
        self.sizer.Add(self.visibility, border=5, flag=wx.ALL | wx.EXPAND)
        self.sizer.Add(wx.StaticLine(self, -1, style=wx.LI_HORIZONTAL),
                       flag=wx.ALL | wx.EXPAND)
        self.sizer.Add(self.description, border=10, flag=wx.ALL | wx.EXPAND)

        self.sizer.Add(wx.StaticLine(self, -1, style=wx.LI_HORIZONTAL),
                       flag=wx.ALL | wx.EXPAND)
        self.sizer.Add(self.syncButton,
                       flag=wx.ALL | wx.RIGHT, border=5)
        self.sizer.Add(self.syncPanel, border=5, proportion=1,
                       flag=wx.ALL | wx.RIGHT | wx.EXPAND)

        if self.project:
            self.setProject(self.project)
            self.syncPanel.setStatus(_translate("Ready to sync"))
        else:
            self.syncPanel.setStatus(
                    _translate("This file doesn't belong to a project yet"))

        self.SetAutoLayout(True)
        self.SetSizerAndFit(self.sizer)
        self.SetupScrolling()
        self.Bind(wx.EVT_SIZE, self.onResize)


    def setProject(self, project, localRoot=''):
        if not isinstance(project, pavlovia.PavloviaProject):
            project = pavlovia.getCurrentSession().getProject(project)
        if project is None:
            return  # we're done
        self.project = project

        if not self.noTitle:
            # use the id (namespace/name) but give space around /
            self.title.SetLabel(project.id.replace("/", " / "))

        # url
        self.url.SetLabel(self.project.web_url)
        self.url.SetURL(self.project.web_url)

        # public / private
        if hasattr(project.attributes, 'description') and project.attributes['description']:
            self.description.SetLabel(project.attributes['description'])
        else:
            self.description.SetLabel('')
        if not hasattr(project, 'visibility'):
            visib = _translate("User not logged in!")
        elif project.visibility in ['public', 'internal']:
            visib = "Public"
        else:
            visib = "Private"
        self.visibility.SetLabel(_translate("Visibility: {}").format(visib))

        # do we have a local location?
        localFolder = project.localRoot
        if not localFolder:
            localFolder = _translate("<not yet synced>")
        self.localFolderCtrl.SetLabel(_translate("Local root: {}").format(localFolder))

        # Check permissions: login, fork or sync
        perms = project.permissions

        # we've got the permissions value so use it
        if not pavlovia.getCurrentSession().user:
            self.syncButton.SetLabel(_translate('Log in to sync...'))
        elif not perms or perms < pavlovia.permissions['developer']:
            self.syncButton.SetLabel(_translate('Fork + sync...'))
        else:
            self.syncButton.SetLabel(_translate('Sync...'))
        self.syncButton.Enable(True)  # now we have a project we should enable

        while None in project.tags:
            project.tags.remove(None)
        self.tags.SetLabel(_translate("Tags:") + " " + ", ".join(project.tags))
        # call onResize to get correct wrapping of description box and title
        self.onResize()

    def onResize(self, evt=None):
        if self.project is None:
            return
        w, h = self.GetSize()
        # if it hasn't been created yet then we won't have attributes
        if hasattr(self.project, 'attributes') and self.project.attributes['description'] is not None:
                self.description.SetLabel(self.project.attributes['description'])
                self.description.Wrap(w - 20)
        # noTitle in some uses of the detailsPanel
        if not self.noTitle and 'name' in self.project:
            self.title.SetLabel(self.project.name)
            self.title.Wrap(w - 20)
        self.Layout()

    def onSyncButton(self, event):
        if not pavlovia.haveGit:
            noGitWarning(parent=self.parent)
            return 0

        if self.project is None:
            raise AttributeError("User pressed the sync button with no "
                                 "current project existing.")

        # log in first if needed
        if not pavlovia.getCurrentSession().user:
            logInPavlovia(parent=self.parent)
            return

        # fork first if needed
        perms = self.project.permissions
        if not perms or perms < pavlovia.permissions['developer']:
            # specifying the group to fork to has no effect so don't use it
            # dlg = ForkDlg(parent=self.parent, project=self.project)
            # if dlg.ShowModal() == wx.ID_CANCEL:
            #     return
            # else:
            #     newGp = dlg.groupField.GetStringSelection()
            #     newName = dlg.nameField.GetValue()
            fork = self.project.forkTo()  # logged-in user
            self.setProject(fork.id)

        # if project.localRoot doesn't exist, or is empty
        if 'localRoot' not in self.project or not self.project.localRoot:
            # we first need to choose a location for the repository
            newPath = setLocalPath(self, self.project)
            if newPath:
                self.localFolderCtrl.SetLabel(
                    label=_translate("Local root: {}").format(newPath))
            self.project.local = newPath
            self.Layout()
            self.Raise()

        self.syncPanel.setStatus(_translate("Synchronizing..."))
        self.project.sync(infoStream=self.syncPanel.infoStream)
        self.parent.Raise()

    def onBrowseLocalFolder(self, evt):
        self.localFolder = setLocalPath(self, self.project)
        if self.localFolder:
            self.localFolderCtrl.SetLabel(
                label=_translate("Local root: {}").format(self.localFolder))
        self.localFolderCtrl.Wrap(self.GetSize().width)
        self.Layout()
        self.parent.Raise()


class ProjectFrame(wx.Dialog):

    def __init__(self, app, parent=None, style=None,
                 pos=wx.DefaultPosition, project=None):
        if style is None:
            style = (wx.DEFAULT_DIALOG_STYLE | wx.CENTER |
                     wx.TAB_TRAVERSAL | wx.RESIZE_BORDER)
        if project:
            title = project.title
        else:
            title = _translate("Project info")
        self.frameType = 'ProjectInfo'
        wx.Dialog.__init__(self, parent, -1, title=title, style=style,
                           size=(700, 500), pos=pos)
        self.app = app
        self.project = project
        self.parent = parent

        self.detailsPanel = DetailsPanel(parent=self, project=self.project)

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.Add(self.detailsPanel, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizerAndFit(self.mainSizer)

        if self.parent:
            self.CenterOnParent()
        self.Layout()

def syncProject(parent, project=None, closeFrameWhenDone=False):
    """A function to sync the current project (if there is one)

    Returns
    -----------
        1 for success
        0 for fail
        -1 for cancel at some point in the process
    """
    if not pavlovia.haveGit:
        noGitWarning(parent)
        return 0

    isCoder = hasattr(parent, 'currentDoc')

    # Test and reject sync from invalid folders
    if isCoder:
        currentPath = os.path.dirname(parent.currentDoc.filename)
    else:
        currentPath = os.path.dirname(parent.filename)

    currentPath = os.path.normcase(os.path.expanduser(currentPath))
    invalidFolders = [os.path.normcase(os.path.expanduser('~/Desktop')),
                      os.path.normcase(os.path.expanduser('~/My Documents'))]

    if currentPath in invalidFolders:
        wx.MessageBox(("You cannot sync projects from:\n\n"
                      "  - Desktop\n"
                      "  - My Documents\n\n"
                      "Please move your project files to another folder, and try again."),
                      "Project Sync Error",
                      wx.ICON_QUESTION | wx.OK)
        return -1

    if not project and "BuilderFrame" in repr(parent):
        # try getting one from the frame
        project = parent.project  # type: pavlovia.PavloviaProject

    if not project:  # ask the user to create one

        # if we're going to create a project we need user to be logged in
        pavSession = pavlovia.getCurrentSession()
        try:
            username = pavSession.user.username
        except:
            username = logInPavlovia(parent)
        if not username:
            return -1  # never logged in

        # create project dialog
        msg = _translate("This file doesn't belong to any existing project.")
        style = wx.OK | wx.CANCEL | wx.CENTER
        dlg = wx.MessageDialog(parent=parent, message=msg, style=style)
        dlg.SetOKLabel(_translate("Create a project"))
        if dlg.ShowModal() == wx.ID_OK:
            if isCoder:
                if parent.currentDoc:
                    localRoot = os.path.dirname(parent.currentDoc.filename)
                else:
                    localRoot = ''
            else:
                localRoot = os.path.dirname(parent.filename)
            # open the project editor (with no project to create one)
            editor = ProjectEditor(parent=parent, localRoot=localRoot)
            if editor.ShowModal() == wx.ID_OK:
                project = editor.project
            else:
                project = None
        else:
            return -1  # user pressed cancel

    if not project:  # we did our best for them. Give up!
        return 0

    # if project.localRoot doesn't exist, or is empty
    if 'localRoot' not in project or not project.localRoot:
        # we first need to choose a location for the repository
        setLocalPath(parent, project)
        parent.Raise()  # make sure that frame is still visible

    #check that the project does exist remotely
    if not project.pavlovia:
        # project didn't exist at Pavlovia (deleted?)
        recreatorDlg = ProjectRecreator(parent=parent, project=project)
        ok = recreatorDlg.ShowModal()
        if ok > 0:
            project = recreatorDlg.project
        else:
            logging.error("Failed to recreate project to sync with")
            return 0

    # a sync will be necessary so set the target to Runner stdout
    parent.app.showRunner()
    syncFrame = parent.app.runner.stdOut

    if project._newRemote:
        # new remote so this will be a first push
        if project.getRepo(forceRefresh=True) is None:
            # no local repo yet so create one
            project.newRepo(syncFrame)
        # add the local files and commit them
        ok = showCommitDialog(parent=parent, project=project,
                              initMsg="First commit",
                              infoStream=syncFrame)
        if ok == -1:  # cancelled
            syncFrame.Destroy()
            return -1
        syncFrame.setStatus("Pushing files to Pavlovia")
        wx.Yield()
        time.sleep(0.001)
        # git push -u origin master
        try:
            project.firstPush(infoStream=syncFrame)
            project._newRemote = False
        except Exception as e:
            closeFrameWhenDone = False
            syncFrame.statusAppend(traceback.format_exc())
    else:
        # existing remote which we should sync (or clone)
        try:
            ok = project.getRepo(syncFrame)
            if not ok:
                closeFrameWhenDone = False
        except Exception as e:
            closeFrameWhenDone = False
            syncFrame.statusAppend(traceback.format_exc())
        # check for anything to commit before pull/push
        outcome = showCommitDialog(parent, project,
                                   infoStream=syncFrame)
        # 0=nothing to do, 1=OK, -1=cancelled
        if outcome == -1:  # user cancelled
            return -1
        try:
            status = project.sync(syncFrame)
            if status == -1:
                syncFrame.statusAppend("Couldn't sync")
        except Exception:  # not yet sure what errors might occur
            # send the error to panel
            syncFrame.statusAppend(traceback.format_exc())
            return 0

    wx.Yield()
    project._lastKnownSync = time.time()
    if closeFrameWhenDone:
        pass

    return 1


class ForkDlg(wx.Dialog):
    """Simple dialog to help choose the location/name of a forked project"""
    # this dialog is working fine, but the API call to fork to a specific
    # namespace doesn't appear to work
    def __init__(self, project, *args, **kwargs):
        wx.Dialog.__init__(self, *args, **kwargs)

        existingName = project.name
        session = pavlovia.getCurrentSession()
        groups = [session.user.username]
        groups.extend(session.listUserGroups())
        msg = wx.StaticText(self, label="Where shall we fork to?")
        groupLbl = wx.StaticText(self, label="Group:")
        self.groupField = wx.Choice(self, choices=groups)
        nameLbl = wx.StaticText(self, label="Project name:")
        self.nameField = wx.TextCtrl(self, value=project.name)

        fieldsSizer = wx.FlexGridSizer(cols=2, rows=2, vgap=5, hgap=5)
        fieldsSizer.AddMany([groupLbl, self.groupField,
                             nameLbl, self.nameField])

        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonSizer.Add(wx.Button(self, id=wx.ID_OK, label="OK"))
        buttonSizer.Add(wx.Button(self, id=wx.ID_CANCEL, label="Cancel"))

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        mainSizer.Add(msg, 1, wx.ALL, 5)
        mainSizer.Add(fieldsSizer, 1, wx.ALL, 5)
        mainSizer.Add(buttonSizer, 1, wx.ALL | wx.ALIGN_RIGHT, 5)

        self.SetSizerAndFit(mainSizer)
        self.Layout()


class ProjectRecreator(wx.Dialog):
    """Use this Dlg to handle the case of a missing (deleted?) remote project
    """

    def __init__(self, project, parent, *args, **kwargs):
        wx.Dialog.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        self.project = project
        existingName = project.name
        msgText = _translate("points to a remote that doesn't exist (deleted?).")
        msgText += (" "+_translate("What shall we do?"))
        msg = wx.StaticText(self, label="{} {}".format(existingName, msgText))
        choices = [_translate("(Re)create a project"),
                   "{} ({})".format(_translate("Point to an different location"),
                                    _translate("not yet supported")),
                   _translate("Forget the local git repository (deletes history keeps files)")]
        self.radioCtrl = wx.RadioBox(self, label='RadioBox', choices=choices,
                                     majorDimension=1)
        self.radioCtrl.EnableItem(1, False)
        self.radioCtrl.EnableItem(2, False)

        mainSizer = wx.BoxSizer(wx.VERTICAL)
        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonSizer.Add(wx.Button(self, id=wx.ID_OK, label=_translate("OK")),
                      1, wx.ALL, 5)
        buttonSizer.Add(wx.Button(self, id=wx.ID_CANCEL, label=_translate("Cancel")),
                      1, wx.ALL, 5)
        mainSizer.Add(msg, 1, wx.ALL, 5)
        mainSizer.Add(self.radioCtrl, 1, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        mainSizer.Add(buttonSizer, 1, wx.ALL | wx.ALIGN_RIGHT, 1)

        self.SetSizer(mainSizer)
        self.Layout()

    def ShowModal(self):
        if wx.Dialog.ShowModal(self) == wx.ID_OK:
            choice = self.radioCtrl.GetSelection()
            if choice == 0:
                editor = ProjectEditor(parent=self.parent,
                                       localRoot=self.project.localRoot)
                if editor.ShowModal() == wx.ID_OK:
                    self.project = editor.project
                    return 1  # success!
                else:
                    return -1  # user cancelled
            elif choice == 1:
                raise NotImplementedError("We don't yet support redirecting "
                                          "your project to a new location.")
            elif choice == 2:
                raise NotImplementedError("Deleting the local git repo is not "
                                          "yet implemented")
        else:
            return -1



"""ioHub.

.. file: ioHub/util/dialogs.py

Copyright (C) 2012-2013 iSolver Software Solutions
Distributed under the terms of the GNU General Public License (GPL version 3 or any later version).

.. moduleauthor:: Sol Simpson <sol@isolver-software.com> + contributors, please see credits section of documentation.
.. fileauthor:: ioHub Team

------------------------------------------------------------------------------------------------------------------------

"""
import os
import wx
import wx.lib.agw.genericmessagedialog as GMD
from pkg_resources import parse_version

class ioHubDialog(object):
    wxapp = None

    def __init__(self, display_index=0):
        self.dialog = None
        self.display_index = display_index
        if ioHubDialog.wxapp is None:
            if parse_version(wx.version()) < parse_version('2.9'):
                ioHubDialog.wxapp = wx.PySimpleApp()
            else:
                ioHubDialog.wxapp = wx.GetApp()
                if ioHubDialog.wxapp is None:
                    ioHubDialog.wxapp = wx.App(False)

    def set_frame_display(self):
        """Centers a wx window on the given Display index."""
        num_displays = wx.Display.GetCount()
        if self.display_index < 0:
            self.display_index = 0
        if self.display_index >= num_displays:
            self.display_index = 0
        x, y, w, h = wx.Display(self.display_index).GetGeometry()
        self.dialog.SetPosition((x, y))
        self.dialog.Center()

    def Destroy(self):
        if self.dialog is not None:
            self.dialog.Destroy()
            self.dialog = None

    def __del__(self):
        self.Destroy()
#
# ProgressBar
#

class ProgressBarDialog(ioHubDialog):
    """wx based progress bar interface."""

    def __init__(
            self,
            dialogTitle='Progress Dialog',
           dialogText='Percent Complete',
            maxValue=100.0,
            display_index=0):
        ioHubDialog.__init__(self, display_index)
        self.dialog = wx.ProgressDialog(
            dialogTitle,
            dialogText,
            maxValue,
            None,
            wx.PD_AUTO_HIDE | wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME |
            wx.PD_ESTIMATED_TIME | wx.PD_REMAINING_TIME)

        self.set_frame_display()

        self.minimumValue = 0.0
        self.maximumValue = maxValue
        self.currentValue = self.minimumValue

    def updateStatus(self, value):
        self.dialog.Update(value)
        self.currentValue = value

    def close(self):
        self.Destroy()

    def getCurrentStatus(self):
        return self.currentValue

#
# MessageDialog
#


class MessageDialog(ioHubDialog):
    YES_NO_BUTTONS = wx.YES_NO
    OK_BUTTON = wx.OK
    CANCEL_BUTTON = wx.CANCEL
    YES_BUTTON = wx.YES
    NO_BUTTON = wx.NO

    INFORMATION_DIALOG = wx.ICON_INFORMATION
    WARNING_DIALOG = wx.ICON_WARNING
    IMPORTANT_DIALOG = wx.ICON_EXCLAMATION
    ERROR_DIALOG = wx.ICON_ERROR
    QUESTION_DIALOG = wx.ICON_QUESTION

    YES_RESULT = wx.ID_YES
    NO_RESULT = wx.ID_NO
    OK_RESULT = wx.ID_OK
    CANCEL_RESULT = wx.ID_CANCEL

    def __init__(self, msg, title=None, showButtons=wx.OK,
                 dialogType=wx.ICON_INFORMATION, allowCancel=True,
                 display_index=0):
        ioHubDialog.__init__(self, display_index)
        if showButtons not in [
                MessageDialog.YES_NO_BUTTONS,
                MessageDialog.OK_BUTTON]:
            raise AttributeError(
                'MessageDialog showButtons arg must be either MessageDialog.YES_NO_BUTTONS or MessageDialog.OK_BUTTON')
        if showButtons == MessageDialog.YES_NO_BUTTONS:
            showButtons |= wx.YES_DEFAULT
        if allowCancel:
            showButtons |= wx.CANCEL

        if dialogType not in [MessageDialog.INFORMATION_DIALOG,
                              MessageDialog.WARNING_DIALOG,
                              MessageDialog.IMPORTANT_DIALOG,
                              MessageDialog.ERROR_DIALOG,
                              MessageDialog.QUESTION_DIALOG]:
            raise AttributeError(
                'MessageDialog dialogType arg must one of MessageDialog.INFORMATION_DIALOG, MessageDialog.WARNING_DIALOG, MessageDialog.IMPORTANT_DIALOG, MessageDialog.ERROR_DIALOG, MessageDialog.QUESTION_DIALOG.')

        if title is None:
            if dialogType == MessageDialog.INFORMATION_DIALOG:
                title = 'For Your Information'
            elif dialogType == MessageDialog.WARNING_DIALOG:
                title = 'Warning'
            elif dialogType == MessageDialog.IMPORTANT_DIALOG:
                title = 'Important Note'
            elif dialogType == MessageDialog.ERROR_DIALOG:
                title = 'Error'
            elif dialogType == MessageDialog.QUESTION_DIALOG:
                title = 'Input Required'

        d = wx.Display(0)
        x, y, w, h = d.GetGeometry()
        d = None

        self.dialog = GMD.GenericMessageDialog(
            None, msg, title, showButtons | dialogType)  # , wrap=int(w/4))
        # TODO Change to own image
        import images
        self.dialog.SetIcon(images.Mondrian.GetIcon())

        self.set_frame_display()

    def show(self):
        result = self.dialog.ShowModal()
        self.Destroy()
        return result

#
# FileChooserDialog
#


class FileDialog(ioHubDialog):
    PYTHON_SCRIPT_FILES = 'Python source (*.py)|*.py'
    EXCEL_FILES = 'Spreadsheets (*.xls)|*.xls'
    IODATA_FILES = 'ioDataStore Files (*.hdf5)|*.hdf5'
    CONFIG_FILES = 'Configuration Files (*.yaml)|*.yaml'
    TEXT_FILES = 'Text Files (*.txt)|*.txt'
    ALL_FILES = 'All Files (*.*)|*.*'
    OK_RESULT = wx.ID_OK
    CANCEL_RESULT = wx.ID_CANCEL

    def __init__(
            self,
            message='Select a File',
            defaultDir=os.getcwd(),
            defaultFile='',
            openFile=True,
            allowMultipleSelections=False,
            allowChangingDirectories=True,
            fileTypes='All files (*.*)|*.*',
            display_index=0):
        ioHubDialog.__init__(self, display_index)
        dstyle = 0

        if openFile is True:
            dstyle = dstyle | wx.FD_OPEN
        if allowMultipleSelections is True:
            dstyle = dstyle | wx.MULTIPLE
        if allowChangingDirectories is True:
            dstyle = dstyle | wx.CHANGE_DIR

        fileTypesCombined = ''
        if isinstance(fileTypes, (list, tuple)):
            for ft in fileTypes:
                fileTypesCombined += ft
                fileTypesCombined += '|'
            fileTypesCombined = fileTypesCombined[:-1]

        self.dialog = wx.FileDialog(
            None, message=message,
            defaultDir=defaultDir,
            defaultFile=defaultFile,
            wildcard=fileTypesCombined,
            style=dstyle
        )

        self.set_frame_display()

    def show(self):
        result = self.dialog.ShowModal()
        selections = self.dialog.GetPaths()
        self.Destroy()
        return result, selections



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import


import time
import weakref
import numpy

from psychopy import visual, core
from psychopy.iohub.util import win32MessagePump

getTime = core.getTime

###########################################
#
# ScreenState Class------------------------------------------------------------
#


class ScreenState(object):
    _currentState = None
    experimentRuntime = None
    window = None

    def __init__(
        self,
        experimentRuntime=None,
        eventTriggers=None,
        timeout=None,
        background_color=(
            255,
            255,
            255)):
        if ScreenState.experimentRuntime is None:
            ScreenState.experimentRuntime = weakref.ref(experimentRuntime)
            ScreenState.window = weakref.ref(experimentRuntime.window)

        w, h = self.experimentRuntime().devices.display.getPixelResolution()
        self._screen_background_fill = visual.Rect(self.window(), w, h,
                                                   lineColor=background_color,
                                                   lineColorSpace='rgb255',
                                                   fillColor=background_color,
                                                   fillColorSpace='rgb255',
                                                   units='pix',
                                                   name='BACKGROUND',
                                                   opacity=1.0,
                                                   interpolate=False)
        self.stim = dict()
        self.stimNames = []

        if isinstance(eventTriggers, Trigger):
            eventTriggers = [eventTriggers, ]
        elif eventTriggers is None:
            eventTriggers = []

        self.event_triggers = eventTriggers
        self._start_time = None
        self.timeout = timeout
        self.dirty = True

    def setScreenColor(self, rgbColor):
        # self.window().setColor(rgbColor,'rgb255')
        self._screen_background_fill.setFillColor(color=rgbColor,
                                                  colorSpace='rgb255')
        self._screen_background_fill.setLineColor(color=rgbColor,
                                                  colorSpace='rgb255')
        self.dirty = True

    def setEventTriggers(self, triggers):
        self.event_triggers = []
        if isinstance(triggers, Trigger):
            triggers = [triggers, ]
        self.event_triggers = triggers

    def addEventTrigger(self, trigger):
        if isinstance(trigger, Trigger):
            self.event_triggers.append(trigger)
        else:
            raise ValueError('Triggers added to a screen state object'
                             'must be of type DeviceEventTrigger.')

    def getStateStartTime(self):
        return self._start_time

    def getEventTriggers(self):
        return self.event_triggers

    def setTimeout(self, timeout):
        self.timeout = timeout

    # switches to screen state (draws and flips)
    # records flip time as start time for timer if timeout has been specified.
    # monitors the device.getEvents function ptrs that are available and if any events are returned,
    # checks the events against the event masks dict provided. If an event matches, it causes method to return
    # then, if no event masks are provided and an event is received, it will cause the method to return regardless
    # of event type for that device.
    # Otherwise method does not reurn until timeout seconds has passed.
    # Returns: [flip_time, time_since_flip, event]
    # all elements but flip_time may be None. All times are in sec.msec
    def switchTo(self, clearEvents=True, msg=None):
        """Switches to the screen state defined by the class instance. The
        screen stim are built and a flip occurs.

        Three conditions can cause the switchTo method to then return,
        based on whether a timeout and / or DeviceEventTriggers
        have been set with the Screen state when switchTo is called. In all cases
        a tuple of three values is returned, some elements of which may be None
        depending on what resulted in the state exit. The three conditions are:

            #. If no timeout or DeviceEventTriggers have been specified with the ScreenState, switchto() returns after the window.flip() with::

                    (stateStartTime, None, None)

               where stateStartTime is the time the call to flip() returned.

            #. If a timeout has been specified, and that amount of time elapses from the startStartTime, then switchTo() returns with::

                    (stateStartTime, stateDuration, None)

               where:

                      * stateStartTime is the time the call to flip() returned.
                      * stateDuration is the time switchTo() returned minus
                      * stateStartTime; so it should be close to the timeout specified. It may be rounded to the next flip() time interval if something in the state is causing the screen to be updated each frame.

            #. If 1 - N DeviceEventTriggers have been set with the ScreenState, they are monitored to determine if any have triggered.
               If a DeviceEventTrigger has triggered, the triggering event and the triggers callback function are retrieved.
               The deviceEventTrigger is then reset, and the callback is called.

        If a callback returns True, the ScreenState is exited, returning (stateStartTime, stateDuration, exitTriggeringEvent), where:

                * **stateStartTime** is the time the call to flip() returned.
                * **stateDuration** is the time switchTo() returned minus stateStartTime; so it should be close to the timeout specified. It may be rounded to the next flip() time interval if something in the state is causing the screen to be updated each frame.
                * **exitTriggeringEvent** is the Device event (in dict form) that caused the ScreenState to exit.

        If the callback returns False, the ScreenState is not exited, and the the timeout period and DeviceEventTriggers cintinue to be checked.

        """
        ER = self.experimentRuntime()
        localClearEvents = ER.hub.clearEvents
        if clearEvents is False:
            localClearEvents = lambda clearEvents: clearEvents is None

        event_triggers = self.event_triggers

        for trigger in event_triggers:
            trigger.resetTrigger()

        lastMsgPumpTime = 0
        self.build()
        self._start_time = self.flip(text=msg)
        endTime = self._start_time + self.timeout
        localClearEvents('all')

        if event_triggers and len(event_triggers) > 0:
            while getTime() + 0.002 < endTime:
                for trigger in event_triggers:
                    if trigger.triggered() is True:

                        event = trigger.getTriggeringEvent()
                        funcToCall, kwargs = trigger.getTriggeredStateCallback()

                        trigger.resetLastTriggeredInfo()

                        if funcToCall:
                            exitState = funcToCall(
                                self._start_time, getTime() - self._start_time, event, **kwargs)
                            if exitState is True:
                                localClearEvents('all')
                                Trigger.clearEventHistory()
                                return (self._start_time,
                                        getTime() - self._start_time, event)
                        break

                Trigger.clearEventHistory()

                tempTime = getTime()
                if tempTime + 0.002 < endTime:
                    time.sleep(0.001)

                    if tempTime - lastMsgPumpTime > 0.5:
                        win32MessagePump()
                        lastMsgPumpTime = tempTime

            localClearEvents('all')
            while getTime() < endTime:
                pass
            return self._start_time, getTime() - self._start_time, None

        elif self.timeout is not None:
            ER.hub.wait(self.timeout - 0.002)
            localClearEvents('all')

            while getTime() < endTime:
                pass

            return self._start_time, getTime() - self._start_time, None

        return self._start_time, None, None

    def build(self):
        self._screen_background_fill.draw()
        for stimName in self.stimNames:
            self.stim[stimName].draw()
        self.dirty = False

    def flip(self, text=None):
        if self.dirty:
            self.build()
        ftime = self.window().flip()
        ScreenState._currentState = self
        if text is not None:
            self.sendMessage(text, ftime)
        return ftime

    def sendMessage(self, text, mtime=None):
        if mtime is None:
            mtime = getTime()
        mtext = text
        try:
            tracker = self.experimentRuntime().getDevice('tracker')
            if tracker is not None and tracker.isConnected() is True:
                mtext = '%s : tracker_time [%.6f]' % (
                    mtext, tracker.trackerSec())
                tracker.sendMessage(mtext)
            else:
                print('----------------------')
                print('Warning: eyetracker is not connected.')
                print('Msg not sent to eyetracker datafile: ')
                print(mtext)
                print('----------------------')
        except Exception:
            pass
        self.experimentRuntime().hub.sendMessageEvent(mtext, sec_time=mtime)

    @classmethod
    def getCurrentScreenState(cls):
        return cls._currentState


class ClearScreen(ScreenState):

    def __init__(self, experimentRuntime, eventTriggers=None, timeout=None,
                 background_color=(255, 255, 255)):
        ScreenState.__init__(self, experimentRuntime, eventTriggers, timeout,
                             background_color)

    def flip(self, text=None):
        if text is None:
            text = 'CLR_SCREEN SYNC: [%s] ' % (text)
        return ScreenState.flip(self, text)


class InstructionScreen(ScreenState):

    def __init__(self, experimentRuntime, text='Default Text',
                 eventTriggers=None, timeout=None, text_color=[0, 0, 0],
                 text_pos=[0, 0], text_height=32,
                 background_color=(255, 255, 255)):

        ScreenState.__init__(self, experimentRuntime, eventTriggers, timeout,
                             background_color)

        l, t, r, b = self.experimentRuntime().devices.display.getBounds()
        self.stim['TEXTLINE'] = visual.TextStim(
            self.window(),
            text=text,
            pos=text_pos,
            height=text_height,
            color=text_color,
            colorSpace='rgb255',
            units='pix',
            wrapWidth=(
                r - l) * .9)
        self.stimNames.append('TEXTLINE')

    def setText(self, text):
        self.stim['TEXTLINE'].setText(text)
        self.dirty = True

    def setTextColor(self, rgbColor):
        self.stim['TEXTLINE'].setColor(rgbColor, 'rgb255')
        self.dirty = True

    def setTextSize(self, size):
        self.stim['TEXTLINE'].setSize(size)
        self.dirty = True

    def setTextPosition(self, pos):
        self.stim['TEXTLINE'].setPos(pos)

    def flip(self, text=''):
        if text is None:
            text = 'INSTRUCT_SCREEN SYNC: [%s] [%s] ' % (
                self.stim['TEXTLINE'].text[0:30], text)
        return ScreenState.flip(self, text)


class ImageScreen(ScreenState):

    def __init__(self, experimentRuntime, imageName, imagePos=(0, 0),
                 imageSize=None, eventTriggers=None, timeout=None,
                 background_color=(255, 255, 255)):

        ScreenState.__init__(self, experimentRuntime, eventTriggers,
                             timeout, background_color)

        w, h = self.experimentRuntime().devices.display.getPixelResolution()

        if imageSize is None:
            from PIL import Image
            imageSize = Image.open(imageName).size

        self.stim['IMAGE'] = visual.ImageStim(
            self.window(),
            image=imageName,
            pos=imagePos,
            size=imageSize,
            name=imageName)

        self.stim['IMAGE'].imageName = imageName
        self.stimNames.append('IMAGE')

    def setImage(self, imageName):
        self.stim['IMAGE'].setImage(imageName)
        self.stim['IMAGE'].imageName = imageName
        self.dirty = True

    def setImageSize(self, size=None):
        if size is None:
            from PIL import Image
            size = Image.open(self.stim['IMAGE'].imageName).size

        self.stim['IMAGE'].setSize(size)
        self.dirty = True

    def setImagePosition(self, pos):
        self.stim['IMAGE'].setPos(pos)

    def flip(self, text=''):
        if text is None:
            text = 'IMAGE_SCREEN SYNC: [%s] [%s] ' % (
                self.stim['IMAGE'].imageName, text)
        return ScreenState.flip(self, text)


# Trigger Classes ---------------------------------------------------

class Trigger(object):
    __slots__ = ['trigger_function', 'user_kwargs', '_last_triggered_event',
                 'repeat_count', 'triggered_count', '_last_triggered_time']

    def __init__(self, trigger_function=lambda a, b, c: True,
                 user_kwargs={}, repeat_count=0):
        self.trigger_function = trigger_function
        self.user_kwargs = user_kwargs
        self._last_triggered_event = None
        self._last_triggered_time = None
        self.repeat_count = repeat_count
        self.triggered_count = 0

    def triggered(self, **kwargs):
        if 0 <= self.repeat_count < self.triggered_count:
            return False
        return True

    def getTriggeringEvent(self):
        return self._last_triggered_event

    def getTriggeringTime(self):
        return self._last_triggered_time

    def getTriggeredStateCallback(self):
        return self.trigger_function, self.user_kwargs

    def resetLastTriggeredInfo(self):
        self._last_triggered_event = None
        self._last_triggered_time = None

    def resetTrigger(self):
        self.resetLastTriggeredInfo()
        self.triggered_count = 0

    @classmethod
    def getEventBuffer(cls, copy=False):
        return {}

    @classmethod
    def clearEventHistory(cls, returncopy=False):
        if returncopy:
            return {}

# Device EventTrigger Class ---------------------------------------------------


class TimeTrigger(Trigger):
    """TimeTrigger's are used by ScreenState objects.

    A TimeTrigger associates a delay from the provided start_time
    parameter to when the classes triggered() method returns True.
    start_time and delay can be sec.msec float, or a callable object
    (that takes no parameters).

    """
    __slots__ = ['startTime', 'delay', '_start_time']

    def __init__(self, start_time, delay, repeat_count=0,
                 trigger_function=lambda a, b, c: True, user_kwargs={}):
        Trigger.__init__(self, trigger_function, user_kwargs, repeat_count)

        self._start_time = start_time

        if start_time is None or not callable(start_time):
            def startTimeFunc():
                if self._start_time is None:
                    self._start_time = getTime()
                return self._start_time
            self.startTime = startTimeFunc
        else:
            self.startTime = start_time

        self.delay = delay
        if not callable(delay):
            def delayFunc():
                return delay
            self.delay = delayFunc

    def triggered(self, **kwargs):
        if Trigger.triggered(self) is False:
            return False

        if self.startTime is None:
            start_time = kwargs.get('start_time')
        else:
            start_time = self.startTime()

        if self.delay is None:
            delay = kwargs.get('delay')
        else:
            delay = self.delay()

        ct = getTime()
        if ct - start_time >= delay:
            self._last_triggered_time = ct
            self._last_triggered_event = ct
            self.triggered_count += 1
            return True
        return False

    def resetTrigger(self):
        self.resetLastTriggeredInfo()
        self.triggered_count = 0
        self._start_time = None


# Device EventTrigger Class ---------------------------------------------------

class DeviceEventTrigger(Trigger):
    """DeviceEventTrigger are used by SCreenState objects.

    A DeviceEventTrigger associates a set of conditions for a
    DeviceEvent that must be met before the classes triggered() method
    returns True.

    """
    _lastEventsByDevice = dict()
    __slots__ = ['device', 'event_type', 'event_attribute_conditions']

    def __init__(self, device, event_type, event_attribute_conditions={},
                 repeat_count=-1, trigger_function=lambda a, b, c: True,
                 user_kwargs={}):
        Trigger.__init__(self, trigger_function, user_kwargs, repeat_count)
        self.device = device
        self.event_type = event_type
        self.event_attribute_conditions = event_attribute_conditions

    def triggered(self, **kwargs):
        if Trigger.triggered(self) is False:
            return False

        events = self.device.getEvents()
        if events is None:
            events = []
        if self.device in self._lastEventsByDevice:
            self._lastEventsByDevice[self.device].extend(events)
        else:
            self._lastEventsByDevice[self.device] = events
        unhandledEvents = self._lastEventsByDevice.get(self.device, [])

        for event in unhandledEvents:
            foundEvent = True
            if event.type != self.event_type:
                foundEvent = False
            else:
                for (attrname,
                     conds) in self.event_attribute_conditions.items():
                    if isinstance(
                            conds, (list, tuple)) and getattr(
                            event, attrname) in conds:
                        # event_value is a list or tuple of possible values
                        # that are OK
                        pass
                    elif getattr(event, attrname) is conds or getattr(event, attrname) == conds:
                        # event_value is a single value
                        pass
                    else:
                        foundEvent = False

            if foundEvent is True:
                self._last_triggered_time = getTime()
                self._last_triggered_event = event
                self.triggered_count += 1
                return True

        return False

    @classmethod
    def getEventBuffer(cls, copy=False):
        if copy:
            return dict(cls._lastEventsByDevice)
        return cls._lastEventsByDevice

    @classmethod
    def clearEventHistory(cls, returncopy=False):
        eventbuffer = None
        if returncopy:
            eventbuffer = dict(cls._lastEventsByDevice)
        cls._lastEventsByDevice.clear()
        return eventbuffer

    def resetLastTriggeredInfo(self):
        Trigger.resetLastTriggeredInfo(self)
        if self.device in self._lastEventsByDevice:
            del self._lastEventsByDevice[self.device]

##########################################################################

#
# MovementPatterns for Visual Stimuli
#

# sinusoidal movement pattern, based on code posted by
# Michael MacAskill @ psychopy-users:
# https://groups.google.com/forum/psychopy-users

# Changed:
#   - made code class based
#   - does not use current time for target position calc, but time of next
#    retrace start
#   -this tries to ensure a more consistent motion, regardless of when the
#    position update is applied to the stimulus during the retrace interval


class SinusoidalMotion(object):

    def __init__(self,
                 amplitude_xy=(15.0, 0.0),       # max horz,vert excursion
                 peak_velocity_xy=(10.0, 10.0),  # deg/s peak velocity x , y
                 phase_xy=(90.0, 90.0),          # in degrees for x, y
                 display=None,                   # ioHub display class
                 start_time=0.0,  # in seconds , 0.0 means use first flip time
                 ):
        self.amplX, self.amplY = amplitude_xy
        self.peakVelX, self.peakVelY = peak_velocity_xy
        self.phaseX, self.phaseY = numpy.deg2rad(
            phase_xy[0]), numpy.deg2rad(phase_xy[1])
        self.startTime = start_time
        self.lastPositionTime = None
        self.nextFlipTimeEstimate = None

        self.reportedRetraceInterval = display.getRetraceInterval()

        # calculate the omega constants needed for the simple
        # harmonic motion equations:

        self.wX = 0.0
        if self.amplX != 0.0:
            self.freqX = self.peakVelX / (-2.0 * self.amplX * numpy.pi)
            self.wX = 2.0 * numpy.pi * self.freqX

        self.wY = 0.0
        if self.amplY != 0:
            self.freqY = self.peakVelY / (-2.0 * self.amplY * numpy.pi)
            self.wY = 2.0 * numpy.pi * self.freqY

    def getPos(self):
        t = 0.0
        if self.lastPositionTime:
            nextFlipTimeEstimate = self.lastPositionTime + self.reportedRetraceInterval
            while nextFlipTimeEstimate < getTime():
                nextFlipTimeEstimate += self.reportedRetraceInterval
            self.nextFlipTimeEstimate = nextFlipTimeEstimate
            t = nextFlipTimeEstimate - self.startTime

        self.pos = (self.amplX * numpy.cos(self.wX * t + self.phaseX),
                    self.amplY * numpy.sin(self.wY * t + self.phaseY))

        return self.pos

    def setLastFlipTime(self, t):
        if self.lastPositionTime is None:
            self.startTime = t
        self.lastPositionTime = t



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from past.utils import old_div
import sys
import time
from psychopy import logging
try:
    import ctypes
    import ctypes.util
    importCtypesFailed = False
except Exception:
    importCtypesFailed = True
    logging.debug("rush() not available because import ctypes "
                  "failed in contrib/darwin.py")

# constants
KERN_SUCCESS = 0
kCGLCPSwapInterval = ctypes.c_int(222)
# these defined in thread_policy.h from apple (googleable)
THREAD_STANDARD_POLICY = ctypes.c_int(1)
THREAD_STANDARD_POLICY_COUNT = ctypes.c_int(0)
THREAD_EXTENDED_POLICY = ctypes.c_int(1)
THREAD_EXTENDED_POLICY_COUNT = ctypes.c_int(1)
THREAD_TIME_CONSTRAINT_POLICY = ctypes.c_int(2)
THREAD_TIME_CONSTRAINT_POLICY_COUNT = ctypes.c_int(4)
# these were found in pyglet/window/carbon/constants thanks to Alex Holkner
kCFStringEncodingASCII = 0x0600
kCFStringEncodingUnicode = 0x0100
kCFStringEncodingUTF8 = 0x08000100
kCFNumberLongType = 10
# some data types these can be found in various *.defs
CGDirectDisplayID = ctypes.c_void_p
CGDisplayCount = ctypes.c_uint32
CGTableCount = ctypes.c_uint32
CGDisplayCoord = ctypes.c_int32
CGByteValue = ctypes.c_ubyte
CGOpenGLDisplayMask = ctypes.c_uint32
CGRefreshRate = ctypes.c_double
CGCaptureOptions = ctypes.c_uint32
integer_t = ctypes.c_int32
natural_t = ctypes.c_uint32
thread_flavor_t = ctypes.c_int32  # in mach_types.defs
thread_info_t = integer_t * 12  # in mach_types.defs
thread_policy_flavor_t = natural_t  # in mach_types.defs
thread_policy_t = integer_t * 16  # in mach_types.defs
# for use with sysctl()
CTL_HW = ctypes.c_int(6)  # /* generic cpu/io */
HW_BUS_FREQ = ctypes.c_int(14)

# could use carbon instead?
cocoa = ctypes.cdll.LoadLibrary(ctypes.util.find_library("Cocoa"))

# not needed - all the functions seem to be in cocoa:
# mach = ctypes.cdll.LoadLibrary(ctypes.util.find_library("libm"))
# ogl = ctypes.cdll.LoadLibrary(ctypes.util.find_library("OpenGL"))


def _create_cfstring(text):
    # some string parameters need to be converted to SFStrings
    if importCtypesFailed:
        return False

    return cocoa.CFStringCreateWithCString(ctypes.c_void_p(),
                                           text.encode('utf8'),
                                           kCFStringEncodingUTF8)

if importCtypesFailed == False:
    class _timeConstraintThreadPolicy(ctypes.Structure):
        _fields_ = [('period', ctypes.c_uint),  # HZ/160
                    ('computation', ctypes.c_uint),  # HZ/3300
                    ('constrain', ctypes.c_uint),  # HZ/2200
                    ('preemptible', ctypes.c_int)]


def syncSwapBuffers(n):
    """syncSwapBuffers(n)
    if n==1 then buffers will sync, otherwise sync will be turned off.
    """
    try:
        # set v to 1 to enable vsync, 0 to disable vsync
        v = ctypes.c_int(n)
        # this is the parameter index?!
        cocoa.CGLSetParameter(cocoa.CGLGetCurrentContext(),
                              kCGLCPSwapInterval, ctypes.pointer(v))
    except Exception:
        logging.warning("Unable to set vsync mode. Using driver defaults")


def getBusFreq():
    """Get the frequency of the system bus (HZ).
    """
    if importCtypesFailed:
        return False

    mib = (ctypes.c_int * 2)(CTL_HW, HW_BUS_FREQ)
    val = ctypes.c_int()
    intSize = ctypes.c_int(ctypes.sizeof(val))
    cocoa.sysctl(ctypes.byref(mib), 2, ctypes.byref(
        val), ctypes.byref(intSize), 0, 0)
    return val.value


def rush(value=True, realtime=False):
    """Raise the priority of the current thread / process.

    Win32 and macOS only so far - on linux use os.nice(niceIncrement)

    Set with rush(True) or rush(False).

    realtime arg is not used by osx implementation.

    Beware and don't take priority until after debugging your code
    and ensuring you have a way out (e.g. an escape sequence of
    keys within the display loop). Otherwise you could end up locked
    out and having to reboot!
    """
    if importCtypesFailed:
        return False

    if value:
        bus = getBusFreq()
        extendedPolicy = _timeConstraintThreadPolicy()
        # number of cycles in hz (make higher than frame rate)
        extendedPolicy.period = old_div(bus, 160)
        extendedPolicy.computation = old_div(bus, 320)  # half of that period
        # max period that they should be carried out in
        extendedPolicy.constrain = old_div(bus, 640)
        extendedPolicy.preemptible = 1
        extendedPolicy = getThreadPolicy(getDefault=True,
                                         flavour=THREAD_TIME_CONSTRAINT_POLICY)
        err = cocoa.thread_policy_set(cocoa.mach_thread_self(),
                                      THREAD_TIME_CONSTRAINT_POLICY,
                                      # send the address of the struct
                                      ctypes.byref(extendedPolicy),
                                      THREAD_TIME_CONSTRAINT_POLICY_COUNT)
        if err != KERN_SUCCESS:
            logging.error(
                'Failed to set darwin thread policy, with thread_policy_set')
        else:
            logging.info('Successfully set darwin thread to realtime')
    else:
        # revert to default policy
        extendedPolicy = getThreadPolicy(getDefault=True,
                                         flavour=THREAD_STANDARD_POLICY)
        err = cocoa.thread_policy_set(cocoa.mach_thread_self(),
                                      THREAD_STANDARD_POLICY,
                                      # send the address of the struct
                                      ctypes.byref(extendedPolicy),
                                      THREAD_STANDARD_POLICY_COUNT)
    return True


def getThreadPolicy(getDefault, flavour):
    """Retrieve the current (or default) thread policy.

    `getDefault` should be True or False.
    `flavour` should be 1 (standard) or 2 (realtime). Not implemented.

    Returns a ctypes struct with fields:
           .period
           .computation
           .constrain
           .preemptible

    See http://docs.huihoo.com/darwin/kernel-programming-guide/scheduler/chapter_8_section_4.html
    """
    if importCtypesFailed:
        return False

    extendedPolicy = _timeConstraintThreadPolicy()  # to store the infos
    # we want to retrive actual policy or the default
    getDefault = ctypes.c_int(getDefault)
    err = cocoa.thread_policy_get(cocoa.mach_thread_self(),
                                  THREAD_TIME_CONSTRAINT_POLICY,
                                  # send the address of the policy struct
                                  ctypes.byref(extendedPolicy),
                                  ctypes.byref(
                                      THREAD_TIME_CONSTRAINT_POLICY_COUNT),
                                  ctypes.byref(getDefault))
    return extendedPolicy


def getRush():
    """Determine whether or not we are in rush mode. Returns True/False.
    """
    if importCtypesFailed:
        return None
    policy = getThreadPolicy(getDefault=False,
                             flavour=THREAD_TIME_CONSTRAINT_POLICY)
    default = getThreadPolicy(getDefault=True,
                              flavour=THREAD_TIME_CONSTRAINT_POLICY)
    # by default this is zero, so not zero means we've changed it
    return policy.period != default.period


def getScreens():
    """Get a list of display IDs from cocoa.
    """
    if importCtypesFailed:
        return False

    count = CGDisplayCount()
    cocoa.CGGetActiveDisplayList(0, None, ctypes.byref(count))
    displays = (CGDirectDisplayID * count.value)()
    cocoa.CGGetActiveDisplayList(count.value, displays, ctypes.byref(count))

    return [id for id in displays]  # python list


def getScreen(screen):
    """Select `screen` from getScreens(), or raise if bad value.
    """
    screens = getScreens()
    if screen > len(screens) - 1:
        msg = "Requested refresh rate of screen %i, but only have %i screens."
        raise IndexError(msg % (screen, len(screens)))
    return getScreens()[screen]


def getRefreshRate(screen=0):
    """Return the refresh rate of the given screen (typically screen = 0 or 1)

    NB. If two screens are connected with different refresh rates then the
    rate at which we draw may not reflect the refresh rate of the monitor.
    """
    if importCtypesFailed:
        return False

    scrID = getScreen(screen)
    mode = cocoa.CGDisplayCurrentMode(scrID)
    refreshCF = cocoa.CFDictionaryGetValue(mode,
                                           _create_cfstring('RefreshRate'))
    refresh = ctypes.c_long()
    cocoa.CFNumberGetValue(refreshCF, kCFNumberLongType,
                           ctypes.byref(refresh))
    if refresh.value == 0:
        return 60  # probably an LCD
    else:
        return refresh.value


def getScreenSizePix(screen=0):
    """Return the height and width (in pixels) of the given screen.
    (typically screen is 0 or 1) If no screen is given then screen 0 is used.

    h,w = getScreenSizePix()
    """
    if importCtypesFailed:
        return False
    scrID = getScreen(screen)
    h = cocoa.CGDisplayPixelsHigh(scrID)
    w = cocoa.CGDisplayPixelsWide(scrID)
    return [h, w]


def waitForVBL(screen=0, nFrames=1):
    """DEPRECATED: the code for doing this is now smaller and cross-platform
    so is included in visual.Window.flip()

    This version is based on detecting the display beam position. It may give
    unpredictable results for an LCD.
    """
    if importCtypesFailed:
        return False

    scrID = getScreen(screen)
    framePeriod = old_div(1.0, getRefreshRate(screen))
    if screen > 0:  # got multiple screens, check if they have same rate
        mainFramePeriod = old_div(1.0, getRefreshRate(0))
        if mainFramePeriod != framePeriod:
            # CGDisplayBeamPosition is unpredictable in this case - usually
            # synced to the first monitor, but maybe better if 2 gfx cards?
            msg = ("You are trying to wait for blanking on a secondary "
                   "monitor that has a different refresh rate to your "
                   "primary monitor. This is not recommended (likely to "
                   "reduce your frame rate to the primary monitor).")
            logging.warning(msg)
    # when we're in a VBL the current beam position is greater than
    # the screen height (for up to ~30 lines)
    top = getScreenSizePix(screen)[0]
    if cocoa.CGDisplayBeamPosition(scrID) > top:
        nFrames += 1  # we're in a VBL already, wait for one more
    while nFrames > 0:
        beamPos = cocoa.CGDisplayBeamPosition(scrID)  # get current pos
        # choose how long to wait
        # we have at least 5ms to go so can wait for 1ms
        while framePeriod * (top - beamPos) / top > 0.005:
            # time.sleep(0.0001)#actually it seems that time.sleep() waits too
            # long on macOS
            beamPos = cocoa.CGDisplayBeamPosition(scrID)  # get current pos
        # now near top so poll continuously
        while beamPos < top:
            beamPos = cocoa.CGDisplayBeamPosition(scrID)  # get current pos
        # if this was not the last frame, then wait until start of next
        # frame before continuing so that we don't detect the VBL again.
        # If this was the last frame then get back to script asap
        if nFrames > 1:
            while beamPos >= top:
                beamPos = cocoa.CGDisplayBeamPosition(scrID)
        nFrames -= 1


def sendStayAwake():
    """Sends a signal to your system to indicate that the computer is in
    use and should not sleep. This should be sent periodically, but
    PsychoPy will send the signal by default on each screen refresh.

    Added: v1.79.00

    Currently supported on: windows, macOS
    """
    cocoa.UpdateSystemActivity(0)

#beamPos =  cocoa.CGDisplayBeamPosition(1)
# while beamPos<=1000:
#        beamPos =  cocoa.CGDisplayBeamPosition(1)
#        print(beamPos)
# first=last=time.time()
# print(getRefreshRate(1))
# for nFrames in range(20):
#    waitForVBL(1, nFrames=1)
#    time.sleep(0.005)
#    this=time.time()
#    print(this-first, this-last, 1/(this-last))
#    last=this
# rush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Part of the PsychoPy library
Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
Distributed under the terms of the GNU General Public License (GPL).
"""

from __future__ import absolute_import, print_function

from builtins import str
from os import path
from psychopy.experiment.components import BaseComponent, Param, _translate

__author__ = 'Jon Peirce'

# the absolute path to the folder containing this path
thisFolder = path.abspath(path.dirname(__file__))
iconFile = path.join(thisFolder, 'static.png')
tooltip = _translate('Static: Static screen period (e.g. an ISI). '
                     'Useful for pre-loading stimuli.')
_localized = {'Custom code': _translate('Custom code')}


class StaticComponent(BaseComponent):
    """A Static Component, allowing frame rendering to pause.

    E.g., pause while disk is accessed for loading an image
    """
    # override the categories property below
    # an attribute of the class, determines the section in the components panel
    categories = ['Custom']

    def __init__(self, exp, parentName, name='ISI',
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=0.5,
                 startEstim='', durationEstim=''):
        BaseComponent.__init__(self, exp, parentName, name=name)
        self.updatesList = []  # a list of dicts {compParams, fieldName}
        self.type = 'Static'
        self.url = "http://www.psychopy.org/builder/components/static.html"
        hnt = _translate(
            "Custom code to be run during the static period (after updates)")
        self.params['code'] = Param("", valType='code',
                                    hint=hnt,
                                    label=_localized['Custom code'])
        self.order = ['name']  # make name come first (others don't matter)

        hnt = _translate("How do you want to define your start point?")
        self.params['startType'] = Param(startType, valType='str',
                                         allowedVals=['time (s)', 'frame N'],
                                         hint=hnt)
        hnt = _translate("How do you want to define your end point?")
        _allow = ['duration (s)', 'duration (frames)', 'time (s)', 'frame N']
        self.params['stopType'] = Param(stopType, valType='str',
                                        allowedVals=_allow,  # copy not needed
                                        hint=hnt)
        hnt = _translate("When does the component start?")
        self.params['startVal'] = Param(startVal, valType='code',
                                        allowedTypes=[],
                                        hint=hnt)
        hnt = _translate("When does the component end? (blank is endless)")
        self.params['stopVal'] = Param(stopVal, valType='code',
                                       allowedTypes=[],
                                       updates='constant', allowedUpdates=[],
                                       hint=hnt)
        hnt = _translate("(Optional) expected start (s), purely for "
                         "representing in the timeline")
        self.params['startEstim'] = Param(startEstim, valType='code',
                                          allowedTypes=[],
                                          hint=hnt)
        hnt = _translate("(Optional) expected duration (s), purely for "
                         "representing in the timeline")
        self.params['durationEstim'] = Param(durationEstim, valType='code',
                                             allowedTypes=[],
                                             hint=hnt)

    def addComponentUpdate(self, routine, compName, fieldName):
        self.updatesList.append({'compName': compName,
                                 'fieldName': fieldName,
                                 'routine': routine})

    def remComponentUpdate(self, routine, compName, fieldName):
        # have to do this in a loop rather than a simple remove
        target = {'compName': compName, 'fieldName': fieldName,
                  'routine': routine}
        for item in self.updatesList:
            if item == target:
                self.updatesList.remove(item)

    def writeInitCode(self, buff):
        code = ("%(name)s = clock.StaticPeriod(win=win, "
                "screenHz=expInfo['frameRate'], name='%(name)s')\n")
        buff.writeIndented(code % self.params)

    def writeFrameCode(self, buff):
        self.writeStartTestCode(buff)
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)
        self.writeStopTestCode(buff)

    def writeStartTestCode(self, buff):
        """This will be executed as the final component in the routine
        """
        buff.writeIndented("# *%s* period\n" % (self.params['name']))
        BaseComponent.writeStartTestCode(self, buff)

        if self.params['stopType'].val == 'time (s)':
            durationSecsStr = "%(stopVal)s-t" % (self.params)
        elif self.params['stopType'].val == 'duration (s)':
            durationSecsStr = "%(stopVal)s" % (self.params)
        elif self.params['stopType'].val == 'duration (frames)':
            durationSecsStr = "%(stopVal)s*frameDur" % (self.params)
        elif self.params['stopType'].val == 'frame N':
            durationSecsStr = "(%(stopVal)s-frameN)*frameDur" % (self.params)
        else:
            msg = ("Couldn't deduce end point for startType=%(startType)s, "
                   "stopType=%(stopType)s")
            raise Exception(msg % self.params)
        vals = (self.params['name'], durationSecsStr)
        buff.writeIndented("%s.start(%s)\n" % vals)

    def writeStopTestCode(self, buff):
        """Test whether we need to stop
        """
        code = ("elif %(name)s.status == STARTED:  # one frame should "
                "pass before updating params and completing\n")
        buff.writeIndented(code % self.params)
        buff.setIndentLevel(+1, relative=True)  # entered an if statement
        self.writeParamUpdates(buff)
        code = "%(name)s.complete()  # finish the static period\n"
        buff.writeIndented(code % self.params)
        # Calculate stop time
        if self.params['stopType'].val == 'time (s)':
            code = "%(name)s.tStop = %(stopVal)s  # record stop time\n"
        elif self.params['stopType'].val == 'duration (s)':
            code = "%(name)s.tStop = %(name)s.tStart + %(stopVal)s  # record stop time\n"
        elif self.params['stopType'].val == 'duration (frames)':
            code = "%(name)s.tStop = %(name)s.tStart + %(stopVal)s*frameDur  # record stop time\n"
        elif self.params['stopType'].val == 'frame N':
            code = "%(name)s.tStop = %(stopVal)s*frameDur  # record stop time\n"
        else:
            msg = ("Couldn't deduce end point for startType=%(startType)s, "
                   "stopType=%(stopType)s")
            raise Exception(msg % self.params)
        # Store stop time
        buff.writeIndented(code % self.params)
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)

        # pass  # the clock.StaticPeriod class handles its own stopping

    def writeParamUpdates(self, buff, updateType=None, paramNames=None):
        """Write updates. Unlike most components, which us this method
        to update themselves, the Static Component uses this to update
        *other* components
        """
        if updateType == 'set every repeat':
            return  # the static component doesn't need to change itself
        if len(self.updatesList):
            code = "# updating other components during *%s*\n"
            buff.writeIndented(code % self.params['name'])
            for update in self.updatesList:
                # update = {'compName':compName,'fieldName':fieldName,
                #    'routine':routine}
                compName = update['compName']
                fieldName = update['fieldName']
                routine = self.exp.routines[update['routine']]
                if hasattr(compName, 'params'):
                    prms = compName.params  # it's already a compon so get params
                else:
                    # it's a name so get compon and then get params
                    prms = self.exp.getComponentFromName(str(compName)).params
                self.writeParamUpdate(buff, compName=compName,
                                      paramName=fieldName,
                                      val=prms[fieldName],
                                      updateType=prms[fieldName].updates,
                                      params=prms)
            code = "# component updates done\n"

            # Write custom code
            if self.params['code']:
                code += ("# Adding custom code for {name}\n"
                         "{code}\n".format(name=self.params['name'],
                                           code=self.params['code']))

            buff.writeIndentedLines(code)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of text using pygame
"""

from __future__ import division

from psychopy import visual, core, event
import sys

# Create a window to draw in
win = visual.Window((800, 800), allowGUI=False,
            monitor='testMonitor', units='cm', winType='pygame')
win.recordFrameIntervals = True

if sys.platform == 'win32':
    fancy = 'c:\\windows\\fonts\\brush'  # this will find brush script
    sans = 'arial'  # on windows you can use short names for any system fonts
    serif = 'c:\\windows\\fonts\\timesi.ttf'  # times in (genuine) italic
    comic = 'c:\\windows\\fonts\\comic.ttf'  # comic
elif sys.platform.startswith('linux'):
    # Note: paths are for Debian-based systems and using the fonts from
    #       ttf-sil-gentium-basic and ttf-dejavu-core from Debian main and
    #       use ttf-mscorefonts-installer from contrib for Comic font
    fancy = '/usr/share/fonts/truetype/ttf-sil-gentium-basic/GenBkBasI.ttf'
    sans = '/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf'
    serif = '/usr/share/fonts/truetype/ttf-dejavu/DejaVuSerif.ttf'
    comic = '/usr/share/fonts/truetype/msttcorefonts/Comic_Sans_MS.ttf'
else:
    # Note that you must have a * .ttf font matching these names/paths
    # download ttf fonts free http://www.webpagepublicity.com/free-fonts.html
    fancy = '/Library/Fonts/Zapfino'
    sans = '/Library/Fonts/Microsoft/Gill Sans MT.ttf'
    serif = '/Library/Fonts/Times New Roman.ttf'
    comic = '/Library/Fonts/Comic Sans MS.ttf'

# Initialize some stimuli
fpsText = visual.TextStim(win,
    units='norm', height=0.1,
    pos=(-0.98, -0.98), text='starting...',
    font=sans,
    anchorVert='bottom', anchorHoriz='left',
    color=[1, -1, -1])
rotating = visual.TextStim(win, text="Fonts rotate!", pos=(0, 0),
    color=[-1, -1, 1], units='deg',
    ori=0, height=1.0,
    font=comic)
unicodeStuff = visual.TextStim(win,
    # you can find the unicode character values online
    text=u"unicode (eg \u03A8 \u040A \u03A3)",
    color=-1, font=serif, pos=(0, 3),
    height=1)
psychopyTxt = visual.TextStim(win, color=1,
    text=u"PsychoPy \u00A9Jon Peirce",
    units='norm', height=0.1,
    pos=[0.95, 0.95],
    alignText='right', anchorHoriz='right', anchorVert='top',
    font=fancy, italic=True)

trialClock = core.Clock()
t = lastFPSupdate = 0
while not event.getKeys():
    t = trialClock.getTime()
    rotating.ori += 1
    rotating.draw()
    unicodeStuff.draw()

    # update the fps every second
    if t - lastFPSupdate > 1:
        fpsText.text = "%i fps" % win.fps()
        lastFPSupdate += 1
    fpsText.draw()
    psychopyTxt.draw()

    # in pygame mouse and key events share one buffer. Need to clear them
    # in case the large number of mouse events fill the buffer preventing keys to be seen
    event.clearEvents('mouse')  # only really needed for pygame windows

    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).

from past.types import basestring
import numpy as N
from .. import Device, DeviceEvent, Computer, Device
from ...constants import DeviceConstants, EventConstants

currentSec = Computer.getTime


class Experiment(Device):
    """
    The Experiment class represents a *virtual* device ( the Python run-time
    within the PsychoPy Process ), and is unique in that it is the *client* of
    the ioHub Event Monitoring Framework, but can also generate events
    itself that are registered with the ioHub process.

    The Experiment Device supports the creation of general purpose MessageEvent's,
    which can effectively hold any string up to 128 characters in length.
    Experiment Message events can be sent to the ioHub Server at any time,
    and are useful for creating stimulus onset or offset notifications, or
    other experiment events of interest that should be associated
    with events from other devices for post hoc analysis of the experiments event steam using
    the ioDataStore.

    The Experiment Device also support LogEvents, which result in the log text sent in the event
    being saved in both the PsychoPy logging module file(s) that have been defined, as well
    as in the ioHub DataStore. The ioHub Process itself actually uses the LogEvent
    type to log status and debugging related information to the DataStore and your log files
    if the log level accepts DEBUG messages.
    """
    EVENT_CLASS_NAMES = ['MessageEvent', 'LogEvent']

    DEVICE_TYPE_ID = DeviceConstants.EXPERIMENT
    DEVICE_TYPE_STRING = 'EXPERIMENT'
    __slots__ = [
        'critical',
        'fatal',
        'error',
        'warning',
        'warn',
        'data',
        'exp',
        'info',
        'debug']

    def __init__(self, *args, **kwargs):
        Device.__init__(self, *args, **kwargs['dconfig'])

        self.critical = lambda text, ltime: self.log(
            text, LogEvent.CRITICAL, ltime)
        self.fatal = self.critical
        self.error = lambda text, ltime: self.log(text, LogEvent.ERROR, ltime)
        self.warning = lambda text, ltime: self.log(
            self, text, LogEvent.WARNING, ltime)
        self.warn = self.warning
        self.data = lambda text, ltime: self.log(text, LogEvent.DATA, ltime)
        self.exp = lambda text, ltime: self.log(text, LogEvent.EXP, ltime)
        self.info = lambda text, ltime: self.log(text, LogEvent.INFO, ltime)
        self.debug = lambda text, ltime: self.log(text, LogEvent.DEBUG, ltime)

    def _nativeEventCallback(self, native_event_data):
        if self.isReportingEvents():
            notifiedTime = currentSec()

            if isinstance(native_event_data, tuple):
                native_event_data = list(native_event_data)

            native_event_data[
                DeviceEvent.EVENT_ID_INDEX] = Device._getNextEventID()

            # set logged time of event
            native_event_data[
                DeviceEvent.EVENT_LOGGED_TIME_INDEX] = notifiedTime

            native_event_data[DeviceEvent.EVENT_DELAY_INDEX] = native_event_data[
                DeviceEvent.EVENT_LOGGED_TIME_INDEX] - native_event_data[DeviceEvent.EVENT_DEVICE_TIME_INDEX]

            native_event_data[
                DeviceEvent.EVENT_HUB_TIME_INDEX] = native_event_data[
                DeviceEvent.EVENT_DEVICE_TIME_INDEX]
            self._addNativeEventToBuffer(native_event_data)

            self._last_callback_time = notifiedTime

    def log(self, text, level=None, log_time=None):
        self._nativeEventCallback(
            LogEvent.create(
                text, level, log_time=log_time))

    def _close(self):
        Device._close(self)

######### Experiment Events ###########


class MessageEvent(DeviceEvent):
    """A MessageEvent can be created and sent to the ioHub to record important
    marker times during the experiment; for example, when key display changes
    occur, or when events related to devices not supported by the ioHub have
    happened, or simply information about the experiment you want to store in
    the ioDataStore along with all the other event data.

    Since the PsychoPy Process can access the same time base that is used by
    the ioHub Process, when you create a Message Event you can time stamp it
    at the time of MessageEvent creation, or with the result of a previous call
    to one of the ioHub time related methods. This makes experiment messages
    extremely accurate temporally when related to other events times saved to
    the ioDataSore.

    Event Type ID: EventConstants.MESSAGE
    Event Type String: 'MESSAGE'

    """
    PARENT_DEVICE = Experiment
    EVENT_TYPE_ID = EventConstants.MESSAGE
    EVENT_TYPE_STRING = 'MESSAGE'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    _newDataTypes = [
        ('msg_offset', N.float32),
        ('category','|S32'),
        ('text', '|S128')
    ]
    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: The text attribute is used to hold the actual 'content' of the message.
        #: The text attribute string can not be more than 128 characters in length.
        #: String type
        self.text = None

        #: The category attribute is a 0 - 32 long string used as a 'group' or 'category'
        #: code that can be assigned to messages. The category attribute may be useful
        #: for grouping messages into categories or types when retieving them for analysis
        #: by assigning the same prix string to related Message Event types.
        #: String type.
        self.category = None

        #: The msg_offset attribute can be used in cases where the Experiment Message
        #: Evenet needs to be sent *before* or *after* the time the actual event occurred.
        #: msg offset should be in sec.msec format and in general can be calculated as:
        #:
        #:      msg_offset=actual_event_iohub_time - iohub_message_time
        #:
        #: where actual_event_iohub_time is the time the event occurred that is being
        #: represented by the Message event; and iohub_message_time is either the
        #: time provided to the Experiment Message creation methods to be used as the
        #: Message time stamp, or is the time that the Message Event actually requested the
        #: current time if no message time was provided.
        #: Both times must be read from Computer.getTime() or one of it's method aliases.
        #: Float type.
        self.msg_offset = None

        DeviceEvent.__init__(self, *args, **kwargs)

    @staticmethod
    def _createAsList(
            text,
            category='',
            msg_offset=0.0,
            sec_time=None,
            set_event_id=None):
        csec = currentSec()
        if set_event_id is None:
            set_event_id = Computer.is_iohub_process
        if sec_time is not None:
            csec = sec_time
        event_num = 0
        if set_event_id:
            event_num = Device._getNextEventID()
        return (0, 0, 0, event_num, MessageEvent.EVENT_TYPE_ID,
                csec, 0, 0, 0.0, 0.0, 0, msg_offset, category, text)


class LogEvent(DeviceEvent):
    """A LogEvent creates an entry in the ioHub dataStore logging table. If
    psychopy is available, it also sends the LogEvent to the Experiment Process
    and it is entered into the psychopy.logging module.

    The LogEvent is unique in that instances can be created by the ioHub Server
    or the Experiment Process psychopy script. In either case, the log entry is
    entered in the psychopy.logging module and the DataStore log table
    (if the datastore is enabled).

    It is **critical** that log events created by the psychopy script use the
    core.getTime() method to provide a LogEvent with the time of the logged event,
    or do not specify a time, in which case the logging module uses the chared time base
    by default.

    Event Type ID: EventConstants.LOG
    Event Type String: 'LOG'

    """
    _psychopyAvailable = False
    _levelNames = dict()
    try:
        from psychopy.logging import _levelNames
        _psychopyAvailable = True

        for lln, llv in _levelNames.items():
            if isinstance(lln, basestring):
                _levelNames[lln] = llv
                _levelNames[llv] = lln
    except Exception:
        CRITICAL = 50
        FATAL = CRITICAL
        ERROR = 40
        WARNING = 30
        WARN = WARNING
        DATA = 25
        EXP = 22
        INFO = 20
        DEBUG = 10
        NOTSET = 0

        _levelNames = {
            CRITICAL: 'CRITICAL',
            ERROR: 'ERROR',
            DATA: 'DATA',
            EXP: 'EXP',
            WARNING: 'WARNING',
            INFO: 'INFO',
            DEBUG: 'DEBUG',
            NOTSET: 'NOTSET',
            'CRITICAL': CRITICAL,
            'ERROR': ERROR,
            'DATA': DATA,
            'EXP': EXP,
            'WARN': WARNING,
            'WARNING': WARNING,
            'INFO': INFO,
            'DEBUG': DEBUG,
            'NOTSET': NOTSET
        }

    PARENT_DEVICE = Experiment
    EVENT_TYPE_ID = EventConstants.LOG
    EVENT_TYPE_STRING = 'LOG'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    _newDataTypes = [
        ('log_level', N.uint8),
        ('text', '|S128')
    ]
    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: The text attribute is used to hold the actual 'content' of the message.
        #: The text attribute string can not be more than 128 characters in length.
        #: String type
        self.text = None

        #: The log level to set the log event at. If psychopy is available,
        #: valid log levels match the *predefined* logging levels. Otherwise
        #: the following log levels are available (which match the predefined
        #: psychopy 2.76 logging settings):
        #:
        #:      * CRITICAL = 50
        #:      * FATAL = CRITICAL
        #:      * ERROR = 40
        #:      * WARNING = 30
        #:      * WARN = WARNING
        #:      * DATA = 25
        #:      * EXP = 22
        #:      * INFO = 20
        #:      * DEBUG = 10
        #:      * NOTSET = 0
        #:
        #: The int defined value can be used, or a string version of the
        #: log level name, for example specifying "WARNING" is equal to
        #: specifying a log level of LogEvent.WARNING or 30.
        #: Default: NOTSET
        #: String or Int type.
        self.log_level = None

        DeviceEvent.__init__(self, *args, **kwargs)

    @classmethod
    def create(cls, msg, level=None, log_time=None):
        created_time = currentSec()
        if log_time is None:
            log_time = created_time
        if level is None or level not in cls._levelNames:
            level = cls.DEBUG
        elif isinstance(level, basestring):
            level = cls._levelNames[level]
        return cls._createAsList(msg, level, created_time, log_time)

    @staticmethod
    def _createAsList(text, log_level, created_time, log_time):
        return (0, 0, 0, 0, LogEvent.EVENT_TYPE_ID,
                created_time, 0, 0, 0.0, 0.0, 0, log_level, text)

    @classmethod
    def _convertFields(cls, event_value_list):
        log_level_value_index = cls.CLASS_ATTRIBUTE_NAMES.index('log_level')
        event_value_list[log_level_value_index] = cls._levelNames.get(
            event_value_list[log_level_value_index], 'UNKNOWN')

    @classmethod
    def createEventAsDict(cls, values):
        cls._convertFields(values)
        return dict(zip(cls.CLASS_ATTRIBUTE_NAMES, values))

    @classmethod
    def createEventAsNamedTuple(cls, valueList):
        cls._convertFields(valueList)
        return cls.namedTupleClass(*valueList)

if not hasattr(LogEvent, 'CRITICAL'):
    for lln, llv in LogEvent._levelNames.items():
        if isinstance(lln, basestring):
            setattr(LogEvent, lln, llv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""
Module containing validators for various parameters.
"""
from __future__ import absolute_import, print_function

from past.builtins import basestring
import wx

import psychopy.experiment.utils
from psychopy.localization import _translate
from . import experiment
from .localizedStrings import _localized
from pkg_resources import parse_version

if parse_version(wx.__version__) < parse_version('4.0.0a1'):
    _ValidatorBase = wx.PyValidator
else:
    _ValidatorBase = wx.Validator

from pyglet.window import key


class BaseValidator(_ValidatorBase):
    """
    Component name validator for _BaseParamsDlg class. It depends on access
    to an experiment namespace.

    Validate calls check, which needs to be implemented per class.

    Messages are passed to user as text in nameOklabel.

    @see: _BaseParamsDlg
    """

    def __init__(self):
        super(BaseValidator, self).__init__()

    def Clone(self):
        return self.__class__()

    def Validate(self, parent):
        """
        """
        # we need to find the dialog to which the Validate event belongs
        # (the event might be fired by a sub-panel and won't have builder exp)
        while not hasattr(parent, 'frame'):
            try:
                parent = parent.GetParent()
            except Exception:
                print("Couldn't find the root dialog for this event")
        message, valid = self.check(parent)
        self.updateMessage(parent)
        return valid

    def TransferFromWindow(self):
        return True

    def TransferToWindow(self):
        return True

    def check(self, parent):
        raise NotImplementedError

    def updateMessage(self, parent):
        """Checks the dict of warning messages for the parent and inserts the
        top one
        :param parent: a component params dialog
        :param message: a
        :return:
        """
        warnings = [w for w in list(parent.warningsDict.values()) if w] or ['']
        msg = warnings[0]
        if parent.nameOKlabel and parent.nameOKlabel.GetLabel() != msg:
            parent.nameOKlabel.SetLabel(msg)
            parent.Layout()


class NameValidator(BaseValidator):
    """Validation checks if the value in Name field is a valid Python
    identifier and if it does not clash with existing names.
    """

    def __init__(self):
        super(NameValidator, self).__init__()

    def check(self, parent):
        """checks namespace, return error-msg (str), enable (bool)
        """
        control = self.GetWindow()
        newName = control.GetValue()
        msg, OK = '', True  # until we find otherwise
        if newName == '':
            msg = _translate("Missing name")
            OK = False
        else:
            namespace = parent.frame.exp.namespace
            used = namespace.exists(newName)
            sameAsOldName = bool(newName == parent.params['name'].val)
            if used and not sameAsOldName:
                msg = _translate("That name is in use (by %s). Try another name.") % _translate(used)
                OK = False
            elif not namespace.isValid(newName):  # valid as a var name
                msg = _translate("Name must be alpha-numeric or _, no spaces")
                OK = False
            # warn but allow, chances are good that its actually ok
            elif namespace.isPossiblyDerivable(newName):
                msg = _translate(namespace.isPossiblyDerivable(newName))
                OK = True
        parent.warningsDict['name'] = msg
        return msg, OK


class CodeSnippetValidator(BaseValidator):
    """Validation checks if field value is intended to be Python code.
    If so, check that it is valid Python, and if valid, check whether
    it contains bad identifiers (currently only self-reference).

    @see: _BaseParamsDlg
    """

    def __init__(self, fieldName):
        super(CodeSnippetValidator, self).__init__()
        self.fieldName = fieldName
        try:
            self.displayName = _localized[fieldName]
        except KeyError:
            # should have all _localized[fieldName] from .localizedStrings
            # might as well try to do something useful if fail:
            self.displayName = _translate(fieldName)

    def Clone(self):
        return self.__class__(self.fieldName)

    def check(self, parent):
        """Checks python syntax of code snippets, and for self-reference.

        Note: code snippets simply use existing names in the namespace,
        like from condition-file headers. They do not add to the
        namespace (unlike Name fields).

        Code snippets in param fields will often be user-defined
        vars, especially condition names. Can also be expressions like
        random(1,100). Hard to know what will be problematic.
        But its always the case that self-reference is wrong.
        """
        # first check if there's anything to validate (and return if not)

        def _checkParamUpdates(parent):
            """Checks whether param allows updates. Returns bool."""
            if parent.params[self.fieldName].allowedUpdates is not None:
                # Check for new set with elements common to lists compared - True if any elements are common
                return bool(set(parent.params[self.fieldName].allowedUpdates) & set(allowedUpdates))

        def _highlightParamVal(parent, error=False):
            """Highlights text containing error - defaults to black"""
            try:
                if error:
                    parent.paramCtrls[self.fieldName].valueCtrl.SetForegroundColour("Red")
                else:
                    parent.paramCtrls[self.fieldName].valueCtrl.SetForegroundColour("Black")
            except KeyError:
                pass

        control = self.GetWindow()
        if not hasattr(control, 'GetValue'):
            return '', True
        val = control.GetValue()  # same as parent.params[self.fieldName].val
        if not isinstance(val, basestring):
            return '', True

        field = self.fieldName
        msg, OK = '', True  # until we find otherwise
        _highlightParamVal(parent)
        codeWanted = psychopy.experiment.utils.unescapedDollarSign_re.search(val)
        isCodeField = bool(parent.params[self.fieldName].valType == 'code')
        allowedUpdates = ['set every repeat', 'set every frame']

        # Check if variable incorrectly defined in correct answer
        allKeyBoardKeys = list(key._key_names.values()) + [str(num) for num in range(10)]
        if self.fieldName == 'correctAns' and not val.startswith('$'):
            if ',' in val:  # comma separated
                keyList = val.upper().split(',')
                keyList = [thisKey.replace(' ', '') for thisKey in keyList if len(thisKey) > 0]
            else:  # whitespace separated
                keyList = val.upper().split(' ')
                keyList = [thisKey.replace(' ', '') for thisKey in keyList if len(thisKey) > 0]

            potentialVars = list(set(keyList) - set(allKeyBoardKeys))  # Elements of keyList not in allKeyBoardKeys
            _highlightParamVal(parent, bool(potentialVars))
            if len(potentialVars):
                msg = _translate("It looks like your 'Correct answer' contains a variable - prepend variables with '$' e.g. ${val}")
                msg = msg.format(val=potentialVars[0].lower())

        if codeWanted or isCodeField:
            # get var names from val, check against namespace:
            code = experiment.getCodeFromParamStr(val)
            try:
                names = compile(code, '', 'exec').co_names
            except (SyntaxError, TypeError) as e:
                # empty '' compiles to a syntax error, ignore
                if not code.strip() == '':
                    _highlightParamVal(parent, True)
                    msg = _translate('Python syntax error in field `{}`:  {}')
                    msg = msg.format(self.displayName, code)
                    OK = False
            else:
                # Check whether variable param entered as a constant
                if isCodeField and _checkParamUpdates(parent):
                    if parent.paramCtrls[self.fieldName].getUpdates() not in allowedUpdates:
                        try:
                            eval(code)
                        except NameError as e:
                            _highlightParamVal(parent, True)
                            msg = _translate(f"Looks like your variable '{code}' in '{self.displayName}' should be set to update.")
                            msg = msg.format(code=code, displayName=self.displayName)
                        except SyntaxError as e:
                            msg = ''

                # namespace = parent.frame.exp.namespace
                # parent.params['name'].val is not in namespace for new params
                # and is not fixed as .val until dialog closes. Use getvalue()
                # to handle dynamic changes to Name field:
                if 'name' in parent.paramCtrls:  # some components don't have names
                    parentName = parent.paramCtrls['name'].getValue()
                    for name in names:
                        # `name` means a var name within a compiled code snippet
                        if name == parentName:
                            _highlightParamVal(parent, True)
                            msg = _translate(
                                'Python var `{}` in `{}` is same as Name')
                            msg = msg.format(name, self.displayName)
                            OK = True

                    for newName in names:
                        namespace = parent.frame.exp.namespace
                        if newName in [*namespace.user, *namespace.builder, *namespace.constants]:
                            continue
                        used = namespace.exists(newName)
                        sameAsOldName = bool(newName == parent.params['name'].val)
                        if used and not sameAsOldName:
                            msg = _translate("Variable name $%s is in use (by %s). Try another name.") % (newName, _translate(used))
                            # let the user continue if this is what they intended
                            OK = True

        parent.warningsDict[field] = msg
        return msg, OK



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script demonstrates how to load a ioHub DataStore HDF5 file, read the
session variable data (if provided when experiment was run, otherwise ignored)
and combine it with columns from a Device Event Table, saving the output as a
tab delimited file.

You will need a .HDF5 file saved using ioHub to run use this demo. If you need
an ioHub data file, first run the demos/coder/iohub/delaytest/run.py script
and use the events.hdf5 file that is saved by that demo. 
"""
from __future__ import absolute_import, division, print_function

from builtins import str
import sys,os
import psychopy
from psychopy.core import getTime
import psychopy.iohub
if psychopy.iohub._DATA_STORE_AVAILABLE is False:
    raise ImportError("DataStore module could not be imported. (Likely that pyTables hdf5dll could not be found). Exiting demo...")
    sys.exit(1)

from psychopy.iohub.datastore.util import displayDataFileSelectionDialog,displayEventTableSelectionDialog, ExperimentDataAccessUtility

def writeOutputFileHeader(output_file, session_metadata_columns,log_entry_names):
    """
    Writes the header line at the top of the Log file.
    Currently uses format:

    session_meta_data_cols [session_user_variable_columns] [event_table_cols][3:]

    Session data is associated with each log entry row using the session_id field.
    """
    allcols=session_metadata_columns+log_entry_names+session_uservar_columns
    output_file.write('\t'.join(allcols))
    output_file.write('\n')

def writeDataRow(output_file,session_info,session_uservar_names,event_data):
    """
    Save a row of data to the output file, in tab delimited format. See comment
    for writeOutputFileHeader function for order of saved columns.
    """
    session_data=[str(i) for i in session_info[:-1]]
    session_user_data=[session_info.user_variables[sud_name] for sud_name in session_uservar_names]
    all_data= session_data+session_user_data+[str(e) for e in event_data]+session_user_data
    output_file.write('\t'.join(all_data))
    output_file.write('\n')

if __name__ == '__main__':
    # Select the hdf5 file to process.
    data_file_path= displayDataFileSelectionDialog(psychopy.iohub.module_directory(writeOutputFileHeader))
    if data_file_path is None:
        print("File Selection Cancelled, exiting...")
        sys.exit(0)
    data_file_path = data_file_path[0]
    dpath,dfile=os.path.split(data_file_path)

    # Lets time how long processing takes
    #
    start_time=getTime()

    # Create an instance of the ExperimentDataAccessUtility class
    # for the selected DataStore file. This allows us to access data
    # in the file based on Device Event names and attributes, as well
    # as access the experiment session metadata saved with each session run.
    dataAccessUtil=ExperimentDataAccessUtility(dpath,dfile, experimentCode=None,sessionCodes=[])

    # Get a dict of all event types -> DataStore table info
    #   for the selected DataStore file.
    eventTableMappings=dataAccessUtil.getEventMappingInformation()

    # Get event tables that have data...
    #
    events_with_data=dataAccessUtil.getEventsByType()

    duration=getTime()-start_time
    # Select which event table to output by displaying a list of
    #   Event Class Names that have data available to the user...
    event_class_selection=displayEventTableSelectionDialog("Select Event Type to Save", "Event Type:",
                [eventTableMappings[event_id].class_name.decode('utf-8') for event_id in list(events_with_data.keys())])
    if event_class_selection is None:
        print("Event table Selection Cancelled, exiting...")
        dataAccessUtil.close()
        sys.exit(0)

    start_time=getTime()

    # Lookup the correct event iterator fiven the event class name selected.
    #
    event_iterator_for_output=None
    for event_id, mapping_info in eventTableMappings.items():
        if mapping_info.class_name.decode('utf-8') == event_class_selection:
            event_iterator_for_output=events_with_data[event_id]
            break
    # Read the session metadata table for all sessions saved to the file.
    #
    session_metadata=dataAccessUtil.getSessionMetaData()

    sesion_meta_data_dict=dict()

    # Create a session_id -> session metadata mapping for use during
    # file writing.
    #
    if len(session_metadata):
        session_metadata_columns = list(session_metadata[0]._fields[:-1] )
        session_uservar_columns=list(session_metadata[0].user_variables.keys())
        for s in session_metadata:
           sesion_meta_data_dict[s.session_id]=s

    # Open a file to save the tab delimited output to.
    #
    if isinstance(event_class_selection, str):
        log_file_name="%s.%s.txt"%(dfile[:-5],event_class_selection)
    else:
        log_file_name="%s.%s.txt"%(dfile[:-5],event_class_selection.decode('utf-8'))
        
    with open(log_file_name,'w') as output_file:
        # write column header
        #
        writeOutputFileHeader(output_file,session_metadata_columns,
                              dataAccessUtil.getEventTable(event_class_selection).cols._v_colnames[3:])

        print('Writing Data to %s:\n'%(log_file_name))
        for i,event in enumerate(event_iterator_for_output):
            # write out each row of the event data with session
            # data as prepended columns.....
            #
            writeDataRow(output_file,sesion_meta_data_dict[event['session_id']],
                         session_uservar_columns,event[:][3:])

            if i%100==0:
                sys.stdout.write('.')

    duration=duration+(getTime()-start_time)
    print()
    print('\nOutput Complete. %d Events Saved to %s in %.3f seconds (%.2f events/seconds).\n'%(i,log_file_name,duration,i/duration))
    print('Output saved to same directory as source .hdf5 file.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Describes the Flow of an experiment
"""

from __future__ import absolute_import, print_function
from past.builtins import basestring

from psychopy.experiment.routine import Routine
from psychopy.experiment.loops import LoopTerminator, LoopInitiator


class Flow(list):
    """The flow of the experiment is a list of L{Routine}s, L{LoopInitiator}s
    and L{LoopTerminator}s, that will define the order in which events occur
    """

    def __init__(self, exp):
        list.__init__(self)
        self.exp = exp
        self._currentRoutine = None
        self._loopList = []  # will be used while we write the code
        self._loopController = {'LoopInitiator': [],
                                'LoopTerminator': []}  # Controls whether loop is written

    @property
    def loopDict(self):
        """Creates a tree of the Flow:
        {entry:object, children:list, parent:object}
        :return:
        """
        loopDict = {}
        currentList = []
        loopStack = [currentList]
        for thisEntry in self:
            if thisEntry.getType() == 'LoopInitiator':
                currentList.append(thisEntry.loop) # this loop is child of current
                loopDict[thisEntry.loop] = []  # and is (current) empty list awaiting children
                currentList = loopDict[thisEntry.loop]
                loopStack.append(loopDict[thisEntry.loop])  # update the list of loops (for depth)
            elif thisEntry.getType() == 'LoopTerminator':
                loopStack.pop()
                currentList = loopStack[-1]
            else:
                # routines should be added to current
                currentList.append(thisEntry)
        return loopDict

    def __repr__(self):
        return "psychopy.experiment.Flow(%s)" % (str(list(self)))

    def addLoop(self, loop, startPos, endPos):
        """Adds initiator and terminator objects for the loop
        into the Flow list"""
        self.insert(int(endPos), LoopTerminator(loop))
        self.insert(int(startPos), LoopInitiator(loop))
        self.exp.requirePsychopyLibs(['data'])  # needed for TrialHandlers etc

    def addRoutine(self, newRoutine, pos):
        """Adds the routine to the Flow list"""
        self.insert(int(pos), newRoutine)

    def removeComponent(self, component, id=None):
        """Removes a Loop, LoopTerminator or Routine from the flow

        For a Loop (or initiator or terminator) to be deleted we can simply
        remove the object using normal list syntax. For a Routine there may
        be more than one instance in the Flow, so either choose which one
        by specifying the id, or all instances will be removed (suitable if
        the Routine has been deleted).
        """
        if component.getType() in ['LoopInitiator', 'LoopTerminator']:
            component = component.loop  # and then continue to do the next
        handlers = ('StairHandler', 'TrialHandler', 'MultiStairHandler')
        if component.getType() in handlers:
            # we need to remove the loop's termination points
            toBeRemoved = []
            for comp in self:
                # bad to change the contents of self when looping through self
                if comp.getType() in ['LoopInitiator', 'LoopTerminator']:
                    if comp.loop == component:
                        # self.remove(comp) --> skips over loop terminator if
                        # its an empty loop
                        toBeRemoved.append(comp)
            for comp in toBeRemoved:
                self.remove(comp)
        elif component.getType() == 'Routine':
            if id is None:
                # a Routine may come up multiple times - remove them all
                # self.remove(component)  # cant do this - two empty routines
                # (with diff names) look the same to list comparison
                toBeRemoved = []
                for id, compInFlow in enumerate(self):
                    if (hasattr(compInFlow, 'name') and
                            component.name == compInFlow.name):
                        toBeRemoved.append(id)
                # need to delete from the end backwards or the order changes
                toBeRemoved.reverse()
                for id in toBeRemoved:
                    del self[id]
            else:
                # just delete the single entry we were given (e.g. from
                # right-click in GUI)
                del self[id]


    def integrityCheck(self):
        """Check that the flow makes sense together and check each component"""

        # force monitor to reload for checks (ie. in case monitor has changed)
        self.exp.settings._monitor = None

        # No checks currently made on flow itself

        trailingWhitespace = []
        constWarnings = []
        for entry in self:
            if hasattr(entry, "integrityCheck"):
                entry.integrityCheck()
            # Now check each routine/loop
            # NB each entry is a routine or LoopInitiator/Terminator
            if not isinstance(entry, Routine):
                continue

            # TODO: the following tests of dubiousConstantUpdates should be
            #  moved into the alerts mechanism under the comp.integrityCheck()
            for component in entry:
                # detect and strip trailing whitespace (can cause problems):
                for key in component.params:
                    field = component.params[key]
                    if not hasattr(field, 'label'):
                        continue  # no problem, no warning
                    if (field.label.lower() in ['text', 'customize'] or
                            not field.valType in ('str', 'code')):
                        continue
                    if (isinstance(field.val, basestring) and
                            field.val != field.val.strip()):
                        trailingWhitespace.append(
                            (field.val, key, component, entry))
                        field.val = field.val.strip()
                # detect 'constant update' fields that seem intended to be
                # dynamic:
                for field, key in component._dubiousConstantUpdates():
                    if field:
                        constWarnings.append(
                            (field.val, key, component, entry))

        if trailingWhitespace:
            warnings = []
            msg = '"%s", in Routine %s (%s: %s)'
            for field, key, component, routine in trailingWhitespace:
                vals = (field, routine.params['name'],
                        component.params['name'], key.capitalize())
                warnings.append(msg % vals)
            print('Note: Trailing white-space removed:\n ', end='')
            # non-redundant, order unknown
            print('\n  '.join(list(set(warnings))))
        if constWarnings:
            warnings = []
            msg = '"%s", in Routine %s (%s: %s)'
            for field, key, component, routine in constWarnings:
                vals = (field, routine.params['name'],
                        component.params['name'], key.capitalize())
                warnings.append(msg % vals)
            print('Note: Dynamic code seems intended but updating '
                  'is "constant":\n ', end='')
            # non-redundant, order unknown
            print('\n  '.join(list(set(warnings))))

    def writePreCode(self,script):
        """Write the code that comes before the Window is created
        """
        script.writeIndentedLines("\n# Start Code - component code to be "
                                  "run before the window creation\n")
        for entry in self:
            # NB each entry is a routine or LoopInitiator/Terminator
            self._currentRoutine = entry
            # very few components need writeStartCode:
            if hasattr(entry, 'writePreCode'):
                entry.writePreCode(script)

    def writeStartCode(self, script):
        """Write the code that comes after the Window is created
        """
        script.writeIndentedLines("\n# Start Code - component code to be "
                                  "run after the window creation\n")
        for entry in self:
            # NB each entry is a routine or LoopInitiator/Terminator
            self._currentRoutine = entry
            # very few components need writeStartCode:
            if hasattr(entry, 'writeStartCode'):
                entry.writeStartCode(script)

    def writeBody(self, script):
        """Write the rest of the code
        """
        # writeStartCode and writeInitCode:
        for entry in self:
            # NB each entry is a routine or LoopInitiator/Terminator
            self._currentRoutine = entry
            entry.writeInitCode(script)
        # create clocks (after initialising stimuli)
        code = ("\n# Create some handy timers\n"
                "globalClock = core.Clock()  # to track the "
                "time since experiment started\n"
                "routineTimer = core.CountdownTimer()  # to "
                "track time remaining of each (non-slip) routine \n")
        script.writeIndentedLines(code)
        # run-time code
        for entry in self:
            self._currentRoutine = entry
            entry.writeMainCode(script)
        # tear-down code (very few components need this)
        for entry in self:
            self._currentRoutine = entry
            entry.writeExperimentEndCode(script)


    def writeFlowSchedulerJS(self, script):
        """Initialise each component and then write the per-frame code too
        """

        # handle email for error messages
        if 'email' in self.exp.settings.params and self.exp.settings.params['email'].val:
            code = ("// If there is an error, we should inform the participant and email the experimenter\n"
                    "// note: we use window.onerror rather than a try/catch as the latter\n"
                    "// do not handle so well exceptions thrown asynchronously\n"
                    "/*window.onerror = function(message, source, lineno, colno, error) {\n"
                    "  console.error(error);\n"
                    "  psychoJS.gui.dialog({'error' : error});\n"
                    "  //psychoJS.core.sendErrorToExperimenter(exception);\n"
                    "  // show error stack on console:\n"
                    "  var json = JSON.parse(error);\n"
                    "  console.error(json.stack);\n"
                    "  return true;\n"
                    "}*/\n")
            script.writeIndentedLines(code)

        code = ("// schedule the experiment:\n"
                "psychoJS.schedule(psychoJS.gui.DlgFromDict({\n"
                "  dictionary: expInfo,\n"
                "  title: expName\n}));\n"
                "\n"
                "const flowScheduler = new Scheduler(psychoJS);\n"
                "const dialogCancelScheduler = new Scheduler(psychoJS);\n"
                "psychoJS.scheduleCondition(function() { return (psychoJS.gui.dialogComponent.button === 'OK'); }, flowScheduler, dialogCancelScheduler);\n"
                "\n")
        script.writeIndentedLines(code)

        code = ("// flowScheduler gets run if the participants presses OK\n"
               "flowScheduler.add(updateInfo); // add timeStamp\n"
               "flowScheduler.add(experimentInit);\n")
        script.writeIndentedLines(code)
        loopStack = []
        for thisEntry in self:
            if not loopStack:  # if not currently in a loop
                if thisEntry.getType() == 'LoopInitiator':
                    code = ("const {name}LoopScheduler = new Scheduler(psychoJS);\n"
                            "flowScheduler.add({name}LoopBegin, {name}LoopScheduler);\n"
                            "flowScheduler.add({name}LoopScheduler);\n"
                            "flowScheduler.add({name}LoopEnd);\n"
                            .format(name=thisEntry.loop.params['name'].val))
                    loopStack.append(thisEntry.loop)
                elif thisEntry.getType() == "Routine":
                    code = ("flowScheduler.add({params[name]}RoutineBegin());\n"
                            "flowScheduler.add({params[name]}RoutineEachFrame());\n"
                            "flowScheduler.add({params[name]}RoutineEnd());\n"
                            .format(params=thisEntry.params))
            else:  # we are already in a loop so don't code here just count
                code = ""
                if thisEntry.getType() == 'LoopInitiator':
                    loopStack.append(thisEntry.loop)
                elif thisEntry.getType() == 'LoopTerminator':
                    loopStack.remove(thisEntry.loop)
            script.writeIndentedLines(code)
        # quit when all routines are finished
        script.writeIndented("flowScheduler.add(quitPsychoJS, '', true);\n")
        # handled all the flow entries
        code = ("\n// quit if user presses Cancel in dialog box:\n"
                "dialogCancelScheduler.add(quitPsychoJS, '', false);\n\n")
        script.writeIndentedLines(code)

        # Write resource list
        resourceFiles = set([resource['rel'].replace("\\", "/") for resource in self.exp.getResourceFiles()])
        if self.exp.htmlFolder:
            resourceFolderStr = "resources/"
        else:
            resourceFolderStr = ""
        script.writeIndented("psychoJS.start({\n")
        script.setIndentLevel(1, relative=True)
        script.writeIndentedLines("expName: expName,\n"
                                  "expInfo: expInfo,\n")
        # if we have an html folder then we moved files there so just use that
        # if not, then we'll need to list all known resource files
        if not self.exp.htmlFolder:
            script.writeIndentedLines("resources: [\n")
            script.setIndentLevel(1, relative=True)
            code = ""
            for idx, resource in enumerate(resourceFiles):
                temp = "{{'name': '{0}', 'path': '{1}{0}'}}".format(resource, resourceFolderStr)
                code += temp
                if idx != (len(resourceFiles)-1):
                    code += ",\n"  # Trailing comma
            script.writeIndentedLines(code)
            script.setIndentLevel(-1, relative=True)
            script.writeIndented("]\n")
            script.setIndentLevel(-1, relative=True)
        script.writeIndented("});\n\n")

    def writeLoopHandlerJS(self, script, modular):
        """
        Function for setting up handler to look after randomisation of conditions etc
        """
        # Then on the flow we need only the Loop Init/terminate
        for entry in self:
            loopType = entry.getType()  # Get type i.e., routine or loop
            if loopType in self._loopController:
                loopName = entry.loop.params['name'].val  # Get loop name
                if loopName not in self._loopController[loopType]:  # Write if not already written
                    entry.writeMainCodeJS(script, modular)  # will either be function trialsBegin() or trialsEnd()
                    self._loopController[loopType].append(loopName)

    def _resetLoopController(self):
        """Resets _loopController so loops are written on each call to write script"""
        self._loopController = {'LoopInitiator': [],
                                'LoopTerminator': []}  # Controls whether loop is written



from builtins import object
from psychopy.visual import Window, CustomMouse, TextStim
import pytest
import pyglet

# currently just a placeholder for better coverage
# checks for syntax errors not proper function: flip, opacity, pos, etc
#
# py.test -k custommouse --cov-report term-missing --cov visual/custommouse.py

@pytest.mark.custommouse
class Test_Custommouse(object):
    @classmethod
    def setup_class(self):
        self.win = Window([128,256])
        self.winpix = Window([128,256], units='pix', autoLog=False)
    @classmethod
    def teardown_class(self):
        for win in [self.win, self.winpix]:
            win.close()

    def test_init(self):
        #for win in [self.win, self.winpix]:
        m = CustomMouse(self.win, showLimitBox=True, autoLog=False)
        assert (m.leftLimit, m.topLimit, m.rightLimit, m.bottomLimit) == (-1, 1, 0.99, -0.98)
        assert m.visible == True
        assert m.showLimitBox == True
        assert m.clickOnUp==False
        m.getPos()
        m.draw()
        m.clickOnUp = m.wasDown = True
        m.isDownNow = False
        m.draw()
        m.getClicks()
        m.resetClicks()
        m.getVisible()
        m.setVisible(False)
        with pytest.raises(AttributeError):
            m.setPointer('a')
        m.setPointer(TextStim(self.win, text='x'))

        m = CustomMouse(self.winpix, autoLog=False)
        assert (m.leftLimit, m.topLimit, m.rightLimit, m.bottomLimit) == (-64.0, 128.0, 59.0, -118.0)
        assert m.visible == True
        assert m.showLimitBox == m.clickOnUp == False
        m.getPos()

    def test_limits(self):
        # to-do: test setLimit
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""utility for creating a .csv file from a .psydat file

edit the file name, then run the script
"""
from __future__ import print_function

import os
from psychopy.tools.filetools import fromFile

# EDIT THE NEXT LINE to be your .psydat file, with the correct path:
name = 'fileName.psydat'

file_psydat = os.path.abspath(name)
print("psydat: {0}".format(file_psydat))

# read in the experiment session from the psydat file:
exp = fromFile(file_psydat)

# write out the data in .csv format (comma-separated tabular text):
if file_psydat.endswith('.psydat'):
    file_csv = file_psydat[:-7]
else:
    file_csv = file_psydat
file_csv += '.csv'
exp.saveAsWideText(file_csv)

print('-> csv: {0}'.format(os.path.abspath(file_csv)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""utility classes for the Builder
"""

from __future__ import absolute_import, division, print_function
from pathlib import Path

from pkg_resources import parse_version
import wx

from psychopy import prefs



from psychopy import visual, event, core
from psychopy.hardware import keyboard

showAll = True  # if False then only show codes that are 'wrong'

print("Keyboards:")
for kb in keyboard.getKeyboards():
    print("{}: {}".format(kb['index'], kb['product']))

kb = keyboard.Keyboard()
win = visual.Window()
msg = visual.TextStim(win, 'press a key')
msg.draw()
win.flip()
keyNames = {}
while True:
    txt = ''
    pygKey = event.waitKeys()[0]
    # wait for the key to be raised
    keys = kb.getKeys(waitRelease=False)
    for key in keys:
        keyCode = key.code
        txt += "{}: {} guess:{}\n".format(keyCode, pygKey, key.name)
        keyNames[keyCode]=pygKey
    msg.text = txt
    msg.draw()
    win.flip()
    if pygKey=='escape':
        break

print(keyNames)
#keyboard = 
#
#while True:
#    
